# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ model.py ]
#   Synopsis     [ the linear model ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import torch
import torch.nn as nn
import torch.nn.functional as F

from argparse import Namespace
from s3prl.upstream.mockingjay.model import TransformerEncoder

#########
# MODEL #
#########

class Identity(nn.Module):
    def __init__(self, config, **kwargs):
        super(Identity, self).__init__()
        # simply take mean operator / no additional parameters

    def forward(self, feature, att_mask, head_mask, **kwargs):

        return [feature]

class Mean(nn.Module):

    def __init__(self, out_dim, input_dim):
        super(Mean, self).__init__()
        self.act_fn = nn.ReLU()
        self.linear = nn.Linear(input_dim, out_dim)
        # simply take mean operator / no additional parameters

    def forward(self, feature, att_mask):

        ''' 
        Arguments
            feature - [BxTxD]   Acoustic feature with shape 
            att_mask   - [BxTx1]     Attention Mask logits
        '''
        feature=self.linear(self.act_fn(feature))
        agg_vec_list = []
        for i in range(len(feature)):
            if torch.nonzero(att_mask[i] < 0, as_tuple=False).size(0) == 0:
                length = len(feature[i])
            else:
                length = torch.nonzero(att_mask[i] < 0, as_tuple=False)[0] + 1
            agg_vec=torch.mean(feature[i][:length], dim=0)
            agg_vec_list.append(agg_vec)
        return torch.stack(agg_vec_list)

class SAP(nn.Module):
    ''' Self Attention Pooling module incoporate attention mask'''

    def __init__(self, out_dim, input_dim):
        super(SAP, self).__init__()

        # Setup
        self.act_fn = nn.ReLU()
        self.linear = nn.Linear(input_dim, out_dim)
        self.sap_layer = SelfAttentionPooling(out_dim)
    
    def forward(self, feature, att_mask):

        ''' 
        Arguments
            feature - [BxTxD]   Acoustic feature with shape 
            att_mask   - [BxTx1]     Attention Mask logits
        '''
        #Encode
        feature = self.act_fn(feature)
        feature = self.linear(feature)
        sap_vec = self.sap_layer(feature, att_mask)

        return sap_vec

class SelfAttentionPooling(nn.Module):
    """
    Implementation of SelfAttentionPooling 
    Original Paper: Self-Attention Encoding and Pooling for Speaker Recognition
    https://arxiv.org/pdf/2008.01077v1.pdf
    """
    def __init__(self, input_dim):
        super(SelfAttentionPooling, self).__init__()
        self.W = nn.Linear(input_dim, 1)
    def forward(self, batch_rep, att_mask):
        """
        input:
        batch_rep : size (N, T, H), N: batch size, T: sequence length, H: Hidden dimension
        
        attention_weight:
        att_w : size (N, T, 1)
        
        return:
        utter_rep: size (N, H)
        """
        seq_len = batch_rep.shape[1]
        softmax = nn.functional.softmax
        att_logits = self.W(batch_rep).squeeze(-1)
        att_logits = att_mask + att_logits
        att_w = softmax(att_logits, dim=-1).unsqueeze(-1)
        utter_rep = torch.sum(batch_rep * att_w, dim=1)

        return utter_rep

class Model(nn.Module):
    def __init__(self, input_dim, agg_dim, agg_module, config):
        super(Model, self).__init__()
        
        # agg_module: current support [ "SAP", "Mean" ]
        # init attributes
        self.agg_method = eval(agg_module)(input_dim, agg_dim)
        
        # two standard transformer encoder layer
        self.model= eval(config['module'])(config=Namespace(**config['hparams']),)
        self.head_mask = [None] * config['hparams']['num_hidden_layers']         
    def forward(self, features, att_mask):
        features = self.model(features,att_mask[:,None,None], head_mask=self.head_mask, output_all_encoded_layers=False)
        utterance_vector = self.agg_method(features[0], att_mask)
        
        return utterance_vector

class UtteranceModel(nn.Module):
    def __init__(self, in_feature, out_dim):
        super(UtteranceModel,self).__init__()
        self.linear1 = nn.Linear(in_feature, out_dim)
        self.linear2 = nn.Linear(out_dim,out_dim)
        self.act_fn = nn.ReLU()
    def forward(self, x):
        hid = self.linear1(x)
        hid = self.act_fn(hid)
        hid = self.linear2(hid)
        hid = self.act_fn(hid)

        return hid

class AdMSoftmaxLoss(nn.Module):

    def __init__(self, in_features, out_features, s=30.0, m=0.4):
        '''
        AM Softmax Loss
        '''
        super(AdMSoftmaxLoss, self).__init__()
        self.s = s
        self.m = m
        self.in_features = in_features
        self.out_features = out_features
        self.fc = nn.Linear(in_features, out_features, bias=False)

    def forward(self, x, labels):
        '''
        input shape (N, in_features)
        '''
        assert len(x) == len(labels)
        assert torch.min(labels) >= 0
        assert torch.max(labels) < self.out_features
        
        for W in self.fc.parameters():
            W = F.normalize(W, dim=1)

        x = F.normalize(x, dim=1)

        wf = self.fc(x)
        numerator = self.s * (torch.diagonal(wf.transpose(0, 1)[labels]) - self.m)
        excl = torch.cat([torch.cat((wf[i, :y], wf[i, y+1:])).unsqueeze(0) for i, y in enumerate(labels)], dim=0)
        denominator = torch.exp(numerator) + torch.sum(torch.exp(self.s * excl), dim=1)
        L = numerator - torch.log(denominator)
        return -torch.mean(L)
    
class TDNN(nn.Module):
        
    def __init__(
                    self, 
                    input_dim=23, 
                    output_dim=512,
                    context_size=5,
                    stride=1,
                    dilation=1,
                    batch_norm=True,
                    dropout_p=0.0
                ):
        '''
        TDNN as defined by https://www.danielpovey.com/files/2015_interspeech_multisplice.pdf
        Affine transformation not applied globally to all frames but smaller windows with local context
        batch_norm: True to include batch normalisation after the non linearity
        
        Context size and dilation determine the frames selected
        (although context size is not really defined in the traditional sense)
        For example:
            context size 5 and dilation 1 is equivalent to [-2,-1,0,1,2]
            context size 3 and dilation 2 is equivalent to [-2, 0, 2]
            context size 1 and dilation 1 is equivalent to [0]
        '''
        super(TDNN, self).__init__()
        self.context_size = context_size
        self.stride = stride
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.dilation = dilation
        self.dropout_p = dropout_p
        self.batch_norm = batch_norm
      
        self.kernel = nn.Linear(input_dim*context_size, output_dim)
        self.nonlinearity = nn.ReLU()
        if self.batch_norm:
            self.bn = nn.BatchNorm1d(output_dim)
        if self.dropout_p:
            self.drop = nn.Dropout(p=self.dropout_p)
        
    def forward(self, x):
        '''
        input: size (batch, seq_len, input_features)
        outpu: size (batch, new_seq_len, output_features)
        '''

        _, _, d = x.shape
        assert (d == self.input_dim), 'Input dimension was wrong. Expected ({}), got ({})'.format(self.input_dim, d)
        x = x.unsqueeze(1)

        # Unfold input into smaller temporal contexts
        x = F.unfold(
                        x, 
                        (self.context_size, self.input_dim), 
                        stride=(1,self.input_dim), 
                        dilation=(self.dilation,1)
                    )

        # N, output_dim*context_size, new_t = x.shape
        x = x.transpose(1,2)
        x = self.kernel(x)
        x = self.nonlinearity(x)
        
        if self.dropout_p:
            x = self.drop(x)

        if self.batch_norm:
            x = x.transpose(1,2)
            x = self.bn(x)
            x = x.transpose(1,2)

        return x

class XVector(nn.Module):
    def __init__(self, config, **kwargs):
        super(XVector, self).__init__()
        # simply take mean operator / no additional parameters
        self.module = nn.Sequential(
            TDNN(input_dim=512, output_dim=512, context_size=5, dilation=1),
            TDNN(input_dim=512, output_dim=512, context_size=3, dilation=2),
            TDNN(input_dim=512, output_dim=512, context_size=3, dilation=3),
            TDNN(input_dim=512, output_dim=512, context_size=1, dilation=1),
            TDNN(input_dim=512, output_dim=1500, context_size=1, dilation=1),
        )

    def forward(self, feature, att_mask, head_mask, **kwargs):
        feature=self.module(feature)
        return [feature]



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/baseline/hubconf.py ]
#   Synopsis     [ the baseline torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import os
#-------------#
from .expert import UpstreamExpert as _UpstreamExpert


def baseline_local(model_config, *args, **kwargs):
    """
        Baseline feature
            model_config: PATH
    """
    assert os.path.isfile(model_config)
    return _UpstreamExpert(model_config, *args, **kwargs)


def baseline(*args, **kwargs):
    """
        Baseline feature - Fbank, or Mel-scale spectrogram
    """
    return fbank(*args, **kwargs)


def spectrogram(*args, **kwargs):
    """
        Baseline feature - Linear-scale spectrogram
    """
    kwargs['model_config'] = os.path.join(os.path.dirname(__file__), 'spectrogram.yaml')
    return baseline_local(*args, **kwargs)


def fbank(*args, **kwargs):
    """
        Baseline feature - Fbank, or Mel-scale spectrogram
    """
    kwargs['model_config'] = os.path.join(os.path.dirname(__file__), 'fbank.yaml')
    return baseline_local(*args, **kwargs)


def fbank_no_cmvn(*args, **kwargs):
    """
        Baseline feature - Fbank, or Mel-scale spectrogram
    """
    kwargs['model_config'] = os.path.join(os.path.dirname(__file__), 'fbank_no_cmvn.yaml')
    return baseline_local(*args, **kwargs)


def mfcc(*args, **kwargs):
    """
        Baseline feature - MFCC
    """
    kwargs['model_config'] = os.path.join(os.path.dirname(__file__), 'mfcc.yaml')
    return baseline_local(*args, **kwargs)


def mel(*args, **kwargs):
    """
        Baseline feature - Mel
    """
    kwargs['model_config'] = os.path.join(os.path.dirname(__file__), 'mel.yaml')
    return baseline_local(*args, **kwargs)


def linear(*args, **kwargs):
    """
        Baseline feature - Linear
    """
    kwargs['model_config'] = os.path.join(os.path.dirname(__file__), 'linear.yaml')
    return baseline_local(*args, **kwargs)



import os
import random
import tqdm
import argparse
import pathlib
import shutil
from pathlib import Path
from librosa.util import find_files
from tqdm import trange

def collect_speaker_ids(roots, speaker_num):
    
    all_speaker=[ ]

    all_speaker.extend([f.path for f in os.scandir(roots) if f.is_dir()])
    
    ids = [[speaker.split("/")[-3],speaker.split("/")[-1]] for speaker in all_speaker]

    vox1 = []
    for id in ids:
        if id[0] == roots.split("/")[-2]:
            vox1.append(id[1])


    dev_speaker = random.sample(vox1, k=speaker_num)
    vox1_train = [ids for ids in vox1 if ids not in dev_speaker]
    
    train_speaker = []

    train_speaker.extend(vox1_train)

    return train_speaker, dev_speaker

def construct_dev_speaker_id_txt(dev_speakers,dev_txt_name):
    f = open(dev_txt_name, "w")
    for dev in dev_speakers:
        f.write(dev)
        f.write("\n")
    f.close()
    return


def sample_wavs_and_dump_txt(root,dev_ids, numbers, meta_data_name):
    
    wav_list = []
    count_positive = 0
    print(f"generate {numbers} sample pairs")
    for _ in trange(numbers):
        prob = random.random()
        if (prob > 0.5):
            dev_id_pair = random.sample(dev_ids, 2)

            # sample 2 wavs from different speaker
            sample1 = "/".join(random.choice(find_files(os.path.join(root,dev_id_pair[0]))).split("/")[-3:])
            sample2 = "/".join(random.choice(find_files(os.path.join(root,dev_id_pair[1]))).split("/")[-3:])

            label = "0"

            wav_list.append(" ".join([label, sample1, sample2]))
            
        else:
            dev_id_pair = random.sample(dev_ids, 1)
            
            # sample 2 wavs from same speaker
            sample1 = "/".join(random.choice(find_files(os.path.join(root,dev_id_pair[0]))).split("/")[-3:])
            sample2 = "/".join(random.choice(find_files(os.path.join(root,dev_id_pair[0]))).split("/")[-3:])

            label = "1"
            count_positive +=1

            wav_list.append(" ".join([label, sample1, sample2]))
    print("finish, then dump file ..")
    f = open(meta_data_name,"w")
    for data in wav_list:
        f.write(data+"\n")
    f.close()

    return wav_list


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--seed', default=19941227)
    parser.add_argument('-r', '--root', default="../../librispeech/voxceleb2/dev/wav")
    parser.add_argument('-o', "--output_dir", default="../../librispeech/dev_data")
    parser.add_argument('-n',  '--speaker_num', default=40)
    parser.add_argument('-p',  '--sample_pair', default=20000)
    args = parser.parse_args()

    random.seed(args.seed)
    train_speakers, dev_speakers = collect_speaker_ids(args.root, args.speaker_num)
    construct_dev_speaker_id_txt(dev_speakers, "./downstream/sv_voxceleb1/dev_meta_data/dev_meta_speaker_ids.txt")
    wav_list = sample_wavs_and_dump_txt(args.root, dev_speakers, args.sample_pair, "./downstream/sv_voxceleb1/dev_meta_data/dev_meta_data_voxceleb2.txt")
    
    root = Path(args.root)

    pathlib.Path(args.output_dir).mkdir(parents=True, exist_ok=True)
    new_dir = Path(args.output_dir)
    print(f"create folder {str(new_dir)} ..")
    pathlib.Path(Path(new_dir / "dev" / "wav")).mkdir(parents=True, exist_ok=True)
    print(f'create folder {str(new_dir / "dev" / "wav" )} ..')
    print(f"copy speaker_folder from {args.root}/dev/wav to {args.output_dir}/dev/wav ..")
    for speaker_fold in tqdm.tqdm(dev_speakers):
        shutil.copytree(root / speaker_fold, new_dir / "dev"/ "wav" / speaker_fold, dirs_exist_ok=True)



import os
import sys
import time
import logging
import hashlib
import shutil
import tempfile
from tqdm import tqdm
from urllib.request import urlopen, Request

from pathlib import Path
from filelock import FileLock

logger = logging.getLogger(__name__)


_download_dir = Path.home() / ".cache" / "s3prl" / "download"


def get_dir():
    _download_dir.mkdir(exist_ok=True, parents=True)
    return _download_dir


def set_dir(d):
    global _download_dir
    _download_dir = Path(d)


def _download_url_to_file(url, dst, hash_prefix=None, progress=True):
    """
    This function is not thread-safe. Please ensure only a single
    thread or process can enter this block at the same time
    """

    file_size = None
    req = Request(url, headers={"User-Agent": "torch.hub"})
    u = urlopen(req)
    meta = u.info()
    if hasattr(meta, "getheaders"):
        content_length = meta.getheaders("Content-Length")
    else:
        content_length = meta.get_all("Content-Length")
    if content_length is not None and len(content_length) > 0:
        file_size = int(content_length[0])

    dst = os.path.expanduser(dst)
    dst_dir = os.path.dirname(dst)
    f = tempfile.NamedTemporaryFile(delete=False, dir=dst_dir)

    try:
        if hash_prefix is not None:
            sha256 = hashlib.sha256()

        tqdm.write(f"Downloading: {url}", file=sys.stderr)
        tqdm.write(f"Destination: {dst}", file=sys.stderr)
        with tqdm(
            total=file_size,
            disable=not progress,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            while True:
                buffer = u.read(8192)
                if len(buffer) == 0:
                    break
                f.write(buffer)
                if hash_prefix is not None:
                    sha256.update(buffer)
                pbar.update(len(buffer))

        f.close()
        if hash_prefix is not None:
            digest = sha256.hexdigest()
            if digest[: len(hash_prefix)] != hash_prefix:
                raise RuntimeError(
                    'invalid hash value (expected "{}", got "{}")'.format(
                        hash_prefix, digest
                    )
                )
        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


def _download(filepath: Path, url, refresh: bool, new_enough_secs: float = 2.0):
    """
    If refresh is True, check the latest modfieid time of the filepath.
    If the file is new enough (no older than `new_enough_secs`), than directly use it.
    If the file is older than `new_enough_secs`, than re-download the file.
    This function is useful when multi-processes are all downloading the same large file
    """

    Path(filepath).parent.mkdir(exist_ok=True, parents=True)

    lock_file = Path(str(filepath) + ".lock")
    logger.info(f"Requesting URL: {url}")

    with FileLock(str(lock_file)):
        if not filepath.is_file() or (
            refresh and (time.time() - os.path.getmtime(filepath)) > new_enough_secs
        ):
            _download_url_to_file(url, filepath)

    logger.info(f"Using URL's local file: {filepath}")
    try:
        lock_file.unlink()
    except FileNotFoundError:
        pass


def _urls_to_filepaths(*args, refresh=False, download: bool = True):
    """
    Preprocess the URL specified in *args into local file paths after downloading

    Args:
        Any number of URLs (1 ~ any)

    Return:
        Same number of downloaded file paths
    """

    def _url_to_filepath(url):
        assert isinstance(url, str)
        m = hashlib.sha256()
        m.update(str.encode(url))
        filepath = get_dir() / f"{str(m.hexdigest())}.{Path(url).name}"
        if download:
            _download(filepath, url, refresh=refresh)
        return str(filepath.resolve())

    paths = [_url_to_filepath(url) for url in args]
    return paths if len(paths) > 1 else paths[0]


urls_to_filepaths = _urls_to_filepaths



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/audio_albert/hubconf.py ]
#   Synopsis     [ the audio albert torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def audio_albert_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
        feature_selection (int): -1 (default, the last layer) or an int in range(0, max_layer_num)
    """
    assert os.path.isfile(ckpt)
    return _UpstreamExpert(ckpt, *args, **kwargs)


def audio_albert_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return audio_albert_local(
        _urls_to_filepaths(ckpt, refresh=refresh), *args, **kwargs
    )


def audio_albert(refresh=False, *args, **kwargs):
    """
        The default model
            refresh (bool): whether to download ckpt/config again if existed
    """
    return audio_albert_960hr(refresh=refresh, *args, **kwargs)


###########
# ALIASES #
###########


def audio_albert_960hr(refresh=False, *args, **kwargs):
    """
    The audio albert base model on 960hr
        refresh (bool): whether to download ckpt/config again if existed
    """
    return audio_albert_logMelBase_T_share_AdamW_b32_1m_960hr_drop1(
        refresh=refresh, *args, **kwargs
    )


##########
# 960 HR #
##########


def audio_albert_logMelBase_T_share_AdamW_b32_1m_960hr_drop1(
    refresh=False, *args, **kwargs
):
    """
    Feature: 80-dim log Mel
    Alteration: time
    Optimizer: AdamW
    Batch size: 32
    Total steps: 1M
    Unlabled Speech: 960hr
    """
    kwargs['ckpt'] = 'https://www.dropbox.com/s/3wgynxmod77ha1z/states-1000000.ckpt?dl=1'
    return audio_albert_url(refresh=refresh, *args, **kwargs)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ utils.py ]
#   Synopsis     [ utility functions ]
#   Author       [ Wen-Chin Huang (https://github.com/unilight) ]
#   Copyright    [ Copyright(c), Toda Lab, Nagoya University, Japan ]
"""*********************************************************************************************"""


import fnmatch
import h5py
import librosa
import logging
import numpy as np
import os
import torch

################################################################################

# The following function are based on:
# https://github.com/espnet/espnet/blob/master/espnet/nets/pytorch_backend/nets_utils.py

def make_pad_mask(lengths, xs=None, length_dim=-1):
    """Make mask tensor containing indices of padded part.

    Args:
        lengths (LongTensor or List): Batch of lengths (B,).
        xs (Tensor, optional): The reference tensor.
            If set, masks will be the same shape as this tensor.
        length_dim (int, optional): Dimension indicator of the above tensor.
            See the example.

    Returns:
        Tensor: Mask tensor containing indices of padded part.
                dtype=torch.uint8 in PyTorch 1.2-
                dtype=torch.bool in PyTorch 1.2+ (including 1.2)

    Examples:
        With only lengths.

        >>> lengths = [5, 3, 2]
        >>> make_non_pad_mask(lengths)
        masks = [[0, 0, 0, 0 ,0],
                 [0, 0, 0, 1, 1],
                 [0, 0, 1, 1, 1]]

        With the reference tensor.

        >>> xs = torch.zeros((3, 2, 4))
        >>> make_pad_mask(lengths, xs)
        tensor([[[0, 0, 0, 0],
                 [0, 0, 0, 0]],
                [[0, 0, 0, 1],
                 [0, 0, 0, 1]],
                [[0, 0, 1, 1],
                 [0, 0, 1, 1]]], dtype=torch.uint8)
        >>> xs = torch.zeros((3, 2, 6))
        >>> make_pad_mask(lengths, xs)
        tensor([[[0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1]],
                [[0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1]],
                [[0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1]]], dtype=torch.uint8)

        With the reference tensor and dimension indicator.

        >>> xs = torch.zeros((3, 6, 6))
        >>> make_pad_mask(lengths, xs, 1)
        tensor([[[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [1, 1, 1, 1, 1, 1]],
                [[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1]],
                [[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1]]], dtype=torch.uint8)
        >>> make_pad_mask(lengths, xs, 2)
        tensor([[[0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1],
                 [0, 0, 0, 0, 0, 1]],
                [[0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1],
                 [0, 0, 0, 1, 1, 1]],
                [[0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1],
                 [0, 0, 1, 1, 1, 1]]], dtype=torch.uint8)

    """
    if length_dim == 0:
        raise ValueError("length_dim cannot be 0: {}".format(length_dim))

    if not isinstance(lengths, list):
        lengths = lengths.tolist()
    bs = int(len(lengths))
    if xs is None:
        maxlen = int(max(lengths))
    else:
        maxlen = xs.size(length_dim)

    seq_range = torch.arange(0, maxlen, dtype=torch.int64)
    seq_range_expand = seq_range.unsqueeze(0).expand(bs, maxlen)
    seq_length_expand = seq_range_expand.new(lengths).unsqueeze(-1)
    mask = seq_range_expand >= seq_length_expand

    if xs is not None:
        assert xs.size(0) == bs, (xs.size(0), bs)

        if length_dim < 0:
            length_dim = xs.dim() + length_dim
        # ind = (:, None, ..., None, :, , None, ..., None)
        ind = tuple(
            slice(None) if i in (0, length_dim) else None for i in range(xs.dim())
        )
        mask = mask[ind].expand_as(xs).to(xs.device)
    return mask


def make_non_pad_mask(lengths, xs=None, length_dim=-1):
    """Make mask tensor containing indices of non-padded part.

    Args:
        lengths (LongTensor or List): Batch of lengths (B,).
        xs (Tensor, optional): The reference tensor.
            If set, masks will be the same shape as this tensor.
        length_dim (int, optional): Dimension indicator of the above tensor.
            See the example.

    Returns:
        ByteTensor: mask tensor containing indices of padded part.
                    dtype=torch.uint8 in PyTorch 1.2-
                    dtype=torch.bool in PyTorch 1.2+ (including 1.2)

    Examples:
        With only lengths.

        >>> lengths = [5, 3, 2]
        >>> make_non_pad_mask(lengths)
        masks = [[1, 1, 1, 1 ,1],
                 [1, 1, 1, 0, 0],
                 [1, 1, 0, 0, 0]]

        With the reference tensor.

        >>> xs = torch.zeros((3, 2, 4))
        >>> make_non_pad_mask(lengths, xs)
        tensor([[[1, 1, 1, 1],
                 [1, 1, 1, 1]],
                [[1, 1, 1, 0],
                 [1, 1, 1, 0]],
                [[1, 1, 0, 0],
                 [1, 1, 0, 0]]], dtype=torch.uint8)
        >>> xs = torch.zeros((3, 2, 6))
        >>> make_non_pad_mask(lengths, xs)
        tensor([[[1, 1, 1, 1, 1, 0],
                 [1, 1, 1, 1, 1, 0]],
                [[1, 1, 1, 0, 0, 0],
                 [1, 1, 1, 0, 0, 0]],
                [[1, 1, 0, 0, 0, 0],
                 [1, 1, 0, 0, 0, 0]]], dtype=torch.uint8)

        With the reference tensor and dimension indicator.

        >>> xs = torch.zeros((3, 6, 6))
        >>> make_non_pad_mask(lengths, xs, 1)
        tensor([[[1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [0, 0, 0, 0, 0, 0]],
                [[1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0]],
                [[1, 1, 1, 1, 1, 1],
                 [1, 1, 1, 1, 1, 1],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0]]], dtype=torch.uint8)
        >>> make_non_pad_mask(lengths, xs, 2)
        tensor([[[1, 1, 1, 1, 1, 0],
                 [1, 1, 1, 1, 1, 0],
                 [1, 1, 1, 1, 1, 0],
                 [1, 1, 1, 1, 1, 0],
                 [1, 1, 1, 1, 1, 0],
                 [1, 1, 1, 1, 1, 0]],
                [[1, 1, 1, 0, 0, 0],
                 [1, 1, 1, 0, 0, 0],
                 [1, 1, 1, 0, 0, 0],
                 [1, 1, 1, 0, 0, 0],
                 [1, 1, 1, 0, 0, 0],
                 [1, 1, 1, 0, 0, 0]],
                [[1, 1, 0, 0, 0, 0],
                 [1, 1, 0, 0, 0, 0],
                 [1, 1, 0, 0, 0, 0],
                 [1, 1, 0, 0, 0, 0],
                 [1, 1, 0, 0, 0, 0],
                 [1, 1, 0, 0, 0, 0]]], dtype=torch.uint8)

    """
    return ~make_pad_mask(lengths, xs, length_dim)

################################################################################

# The following function are based on:
# https://github.com/kan-bayashi/ParallelWaveGAN/blob/master/parallel_wavegan/utils/utils.py

def find_files(root_dir, query="*.wav", include_root_dir=True):
    files = []
    for root, dirnames, filenames in os.walk(root_dir, followlinks=True):
        for filename in fnmatch.filter(filenames, query):
            files.append(os.path.join(root, filename))
    if not include_root_dir:
        files = [file_.replace(root_dir + "/", "") for file_ in files]

    return files

def read_hdf5(hdf5_name, hdf5_path):
    """Read hdf5 dataset.

    Args:
        hdf5_name (str): Filename of hdf5 file.
        hdf5_path (str): Dataset name in hdf5 file.

    Return:
        any: Dataset values.

    """
    if not os.path.exists(hdf5_name):
        logging.error(f"There is no such a hdf5 file ({hdf5_name}).")
        sys.exit(1)

    hdf5_file = h5py.File(hdf5_name, "r")

    if hdf5_path not in hdf5_file:
        logging.error(f"There is no such a data in hdf5 file. ({hdf5_path})")
        sys.exit(1)

    hdf5_data = hdf5_file[hdf5_path][()]
    hdf5_file.close()

    return hdf5_data


def write_hdf5(hdf5_name, hdf5_path, write_data, is_overwrite=True):
    """Write dataset to hdf5.

    Args:
        hdf5_name (str): Hdf5 dataset filename.
        hdf5_path (str): Dataset path in hdf5.
        write_data (ndarray): Data to write.
        is_overwrite (bool): Whether to overwrite dataset.

    """
    # convert to numpy array
    write_data = np.array(write_data)

    # check folder existence
    folder_name, _ = os.path.split(hdf5_name)
    if not os.path.exists(folder_name) and len(folder_name) != 0:
        os.makedirs(folder_name)

    # check hdf5 existence
    if os.path.exists(hdf5_name):
        # if already exists, open with r+ mode
        hdf5_file = h5py.File(hdf5_name, "r+")
        # check dataset existence
        if hdf5_path in hdf5_file:
            if is_overwrite:
                logging.warning("Dataset in hdf5 file already exists. "
                                "recreate dataset in hdf5.")
                hdf5_file.__delitem__(hdf5_path)
            else:
                logging.error("Dataset in hdf5 file already exists. "
                              "if you want to overwrite, please set is_overwrite = True.")
                hdf5_file.close()
                sys.exit(1)
    else:
        # if not exists, open with w mode
        hdf5_file = h5py.File(hdf5_name, "w")

    # write data to hdf5
    hdf5_file.create_dataset(hdf5_path, data=write_data)
    hdf5_file.flush()
    hdf5_file.close()

################################################################################

# The following function are based on:
# https://github.com/espnet/espnet/blob/master/utils/convert_fbank_to_wav.py

EPS = 1e-10

def logmelspc_to_linearspc(lmspc, fs, n_mels, n_fft, fmin=None, fmax=None):
    """Convert log Mel filterbank to linear spectrogram.

    Args:
        lmspc (ndarray): Log Mel filterbank (T, n_mels).
        fs (int): Sampling frequency.
        n_mels (int): Number of mel basis.
        n_fft (int): Number of FFT points.
        f_min (int, optional): Minimum frequency to analyze.
        f_max (int, optional): Maximum frequency to analyze.

    Returns:
        ndarray: Linear spectrogram (T, n_fft // 2 + 1).

    """
    assert lmspc.shape[1] == n_mels
    fmin = 0 if fmin is None else fmin
    fmax = fs / 2 if fmax is None else fmax
    mspc = np.power(10.0, lmspc)
    mel_basis = librosa.filters.mel(fs, n_fft, n_mels, fmin, fmax)
    inv_mel_basis = np.linalg.pinv(mel_basis)
    spc = np.maximum(EPS, np.dot(inv_mel_basis, mspc.T).T)

    return spc


def griffin_lim(spc, n_fft, n_shift, win_length, window="hann", n_iters=100):
    """Convert linear spectrogram into waveform using Griffin-Lim.

    Args:
        spc (ndarray): Linear spectrogram (T, n_fft // 2 + 1).
        n_fft (int): Number of FFT points.
        n_shift (int): Shift size in points.
        win_length (int): Window length in points.
        window (str, optional): Window function type.
        n_iters (int, optionl): Number of iterations of Griffin-Lim Algorithm.

    Returns:
        ndarray: Reconstructed waveform (N,).

    """
    # assert the size of input linear spectrogram
    assert spc.shape[1] == n_fft // 2 + 1

    # use librosa's fast Grriffin-Lim algorithm
    spc = np.abs(spc.T)
    y = librosa.griffinlim(
        S=spc,
        n_iter=n_iters,
        hop_length=n_shift,
        win_length=win_length,
        window=window,
        center=True if spc.shape[1] > 1 else False,
    )

    return y

################################################################################

# The following function are based on:
# https://github.com/espnet/espnet/blob/master/espnet/transform/spectrogram.py

def stft(
    x, n_fft, n_shift, win_length=None, window="hann", center=True, pad_mode="reflect"
):
    # x: [Time, Channel]
    if x.ndim == 1:
        single_channel = True
        # x: [Time] -> [Time, Channel]
        x = x[:, None]
    else:
        single_channel = False
    x = x.astype(np.float32)

    # FIXME(kamo): librosa.stft can't use multi-channel?
    # x: [Time, Channel, Freq]
    x = np.stack(
        [
            librosa.stft(
                x[:, ch],
                n_fft=n_fft,
                hop_length=n_shift,
                win_length=win_length,
                window=window,
                center=center,
                pad_mode=pad_mode,
            ).T
            for ch in range(x.shape[1])
        ],
        axis=1,
    )

    if single_channel:
        # x: [Time, Channel, Freq] -> [Time, Freq]
        x = x[:, 0]
    return x


def stft2logmelspectrogram(x_stft, fs, n_mels, n_fft, fmin=None, fmax=None, eps=1e-10):
    # x_stft: (Time, Channel, Freq) or (Time, Freq)
    fmin = 0 if fmin is None else fmin
    fmax = fs / 2 if fmax is None else fmax

    # spc: (Time, Channel, Freq) or (Time, Freq)
    spc = np.abs(x_stft)
    # mel_basis: (Mel_freq, Freq)
    mel_basis = librosa.filters.mel(fs, n_fft, n_mels, fmin, fmax)
    # lmspc: (Time, Channel, Mel_freq) or (Time, Mel_freq)
    lmspc = np.log10(np.maximum(eps, np.dot(spc, mel_basis.T)))

    return lmspc


def logmelspectrogram(
    x,
    fs,
    n_mels,
    n_fft,
    n_shift,
    win_length=None,
    window="hann",
    fmin=None,
    fmax=None,
    eps=1e-10,
    pad_mode="reflect",
):
    # stft: (Time, Channel, Freq) or (Time, Freq)
    x_stft = stft(
        x,
        n_fft=n_fft,
        n_shift=n_shift,
        win_length=win_length,
        window=window,
        pad_mode=pad_mode,
    )

    return stft2logmelspectrogram(
        x_stft, fs=fs, n_mels=n_mels, n_fft=n_fft, fmin=fmin, fmax=fmax, eps=eps
    )



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ dataset.py ]
#   Synopsis     [ the speaker diarization dataset ]
#   Source       [ Refactored from https://github.com/hitachi-speech/EEND ]
#   Author       [ Jiatong Shi ]
#   Copyright    [ Copyright(c), Johns Hopkins University ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import io
import os
import random
import subprocess
import sys

# -------------#
import numpy as np
import pandas as pd
import soundfile as sf

# -------------#
import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data.dataset import Dataset

# -------------#
import torchaudio


def _count_frames(data_len, size, step):
    # no padding at edges, last remaining samples are ignored
    return int((data_len - size + step) / step)


def _gen_frame_indices(
    data_length,
    size=2000,
    step=2000,
    use_last_samples=False,
    label_delay=0,
    subsampling=1,
):
    i = -1
    for i in range(_count_frames(data_length, size, step)):
        yield i * step, i * step + size
    if use_last_samples and i * step + size < data_length:
        if data_length - (i + 1) * step - subsampling * label_delay > 0:
            yield (i + 1) * step, data_length


def _gen_chunk_indices(data_len, chunk_size):
    step = chunk_size
    start = 0
    while start < data_len:
        end = min(data_len, start + chunk_size)
        yield start, end
        start += step


#######################
# Diarization Dataset #
#######################
class DiarizationDataset(Dataset):
    def __init__(
        self,
        mode,
        data_dir,
        dtype=np.float32,
        chunk_size=2000,
        frame_shift=256,
        subsampling=1,
        rate=16000,
        input_transform=None,
        use_last_samples=True,
        label_delay=0,
        num_speakers=None,
    ):
        super(DiarizationDataset, self).__init__()

        self.mode = mode
        self.data_dir = data_dir
        self.dtype = dtype
        self.chunk_size = chunk_size
        self.frame_shift = frame_shift
        self.subsampling = subsampling
        self.n_speakers = num_speakers
        self.chunk_indices = [] if mode != "test" else {}
        self.label_delay = label_delay

        self.data = KaldiData(self.data_dir)

        # make chunk indices: filepath, start_frame, end_frame
        for rec in self.data.wavs:
            data_len = int(self.data.reco2dur[rec] * rate / frame_shift)
            data_len = int(data_len / self.subsampling)
            if mode == "test":
                self.chunk_indices[rec] = []
            if mode != "test":
                for st, ed in _gen_frame_indices(
                    data_len,
                    chunk_size,
                    chunk_size,
                    use_last_samples,
                    label_delay=self.label_delay,
                    subsampling=self.subsampling,
                ):
                    self.chunk_indices.append(
                        (rec, st * self.subsampling, ed * self.subsampling)
                    )
            else:
                for st, ed in _gen_chunk_indices(data_len, chunk_size):
                    self.chunk_indices[rec].append(
                        (rec, st * self.subsampling, ed * self.subsampling)
                    )

        if mode != "test":
            print(len(self.chunk_indices), " chunks")
        else:
            self.rec_list = list(self.chunk_indices.keys())
            print(len(self.rec_list), " recordings")

    def __len__(self):
        return (
            len(self.rec_list)
            if type(self.chunk_indices) == dict
            else len(self.chunk_indices)
        )

    def __getitem__(self, i):
        if self.mode != "test":
            rec, st, ed = self.chunk_indices[i]
            Y, T = self._get_labeled_speech(rec, st, ed, self.n_speakers)
            # TODO: add subsampling here
            return Y, T
        else:
            chunks = self.chunk_indices[self.rec_list[i]]
            Ys, Ts = [], []
            for (rec, st, ed) in chunks:
                Y, T = self._get_labeled_speech(rec, st, ed, self.n_speakers)
                Ys.append(Y)
                Ts.append(T)
            return Ys, Ts, self.rec_list[i]

    def _get_labeled_speech(
        self, rec, start, end, n_speakers=None, use_speaker_id=False
    ):
        """Extracts speech chunks and corresponding labels

        Extracts speech chunks and corresponding diarization labels for
        given recording id and start/end times

        Args:
            rec (str): recording id
            start (int): start frame index
            end (int): end frame index
            n_speakers (int): number of speakers
                if None, the value is given from data
        Returns:
            data: speech chunk
                (n_samples)
            T: label
                (n_frmaes, n_speakers)-shaped np.int32 array.
        """
        data, rate = self.data.load_wav(
            rec, start * self.frame_shift, end * self.frame_shift
        )
        frame_num = end - start
        filtered_segments = self.data.segments[rec]
        # filtered_segments = self.data.segments[self.data.segments['rec'] == rec]
        speakers = np.unique(
            [self.data.utt2spk[seg["utt"]] for seg in filtered_segments]
        ).tolist()
        if n_speakers is None:
            n_speakers = len(speakers)
        T = np.zeros((frame_num, n_speakers), dtype=np.int32)

        if use_speaker_id:
            all_speakers = sorted(self.data.spk2utt.keys())
            S = np.zeros((frame_num, len(all_speakers)), dtype=np.int32)

        for seg in filtered_segments:
            speaker_index = speakers.index(self.data.utt2spk[seg["utt"]])
            if use_speaker_id:
                all_speaker_index = all_speakers.index(self.data.utt2spk[seg["utt"]])
            start_frame = np.rint(seg["st"] * rate / self.frame_shift).astype(int)
            end_frame = np.rint(seg["et"] * rate / self.frame_shift).astype(int)
            rel_start = rel_end = None
            if start <= start_frame and start_frame < end:
                rel_start = start_frame - start
            if start < end_frame and end_frame <= end:
                rel_end = end_frame - start
            if rel_start is not None or rel_end is not None:
                T[rel_start:rel_end, speaker_index] = 1
                if use_speaker_id:
                    S[rel_start:rel_end, all_speaker_index] = 1

        if use_speaker_id:
            return data, T, S
        else:
            return data, T

    def collate_fn(self, batch):
        batch_size = len(batch)
        len_list = [len(batch[i][1]) for i in range(batch_size)]
        wav = []
        label = []
        for i in range(batch_size):
            length = len_list[i]
            wav.append(batch[i][0].astype(np.float32))
            label.append(batch[i][1].astype(np.float32))
        length = np.array(len_list)
        return wav, label, length, None

    def collate_fn_rec_infer(self, batch):
        assert len(batch) == 1  # each batch should contain one recording
        chunk_num = len(batch[0][1])
        len_list = [len(batch[0][1][i]) for i in range(chunk_num)]
        wav = []
        label = []
        for i in range(chunk_num):
            length = len_list[i]
            wav.append(batch[0][0][i].astype(np.float32))
            label.append(batch[0][1][i].astype(np.float32))
        length = np.array(len_list)
        rec_id = batch[0][2]
        return wav, label, length, rec_id


#######################
# Kaldi-style Dataset #
#######################
class KaldiData:
    """This class holds data in kaldi-style directory."""

    def __init__(self, data_dir):
        """Load kaldi data directory."""
        self.data_dir = data_dir
        self.segments = self._load_segments_rechash(
            os.path.join(self.data_dir, "segments")
        )
        self.utt2spk = self._load_utt2spk(os.path.join(self.data_dir, "utt2spk"))
        self.wavs = self._load_wav_scp(os.path.join(self.data_dir, "wav.scp"))
        self.reco2dur = self._load_reco2dur(os.path.join(self.data_dir, "reco2dur"))
        self.spk2utt = self._load_spk2utt(os.path.join(self.data_dir, "spk2utt"))

    def load_wav(self, recid, start=0, end=None):
        """Load wavfile given recid, start time and end time."""
        data, rate = self._load_wav(self.wavs[recid], start, end)
        return data, rate

    def _load_segments(self, segments_file):
        """Load segments file as array."""
        if not os.path.exists(segments_file):
            return None
        return np.loadtxt(
            segments_file,
            dtype=[("utt", "object"), ("rec", "object"), ("st", "f"), ("et", "f")],
            ndmin=1,
        )

    def _load_segments_hash(self, segments_file):
        """Load segments file as dict with uttid index."""
        ret = {}
        if not os.path.exists(segments_file):
            return None
        for line in open(segments_file):
            utt, rec, st, et = line.strip().split()
            ret[utt] = (rec, float(st), float(et))
        return ret

    def _load_segments_rechash(self, segments_file):
        """Load segments file as dict with recid index."""
        ret = {}
        if not os.path.exists(segments_file):
            return None
        for line in open(segments_file):
            utt, rec, st, et = line.strip().split()
            if rec not in ret:
                ret[rec] = []
            ret[rec].append({"utt": utt, "st": float(st), "et": float(et)})
        return ret

    def _load_wav_scp(self, wav_scp_file):
        """Return dictionary { rec: wav_rxfilename }."""
        if os.path.exists(wav_scp_file):
            lines = [line.strip().split(None, 1) for line in open(wav_scp_file)]
            return {x[0]: x[1] for x in lines}
        else:
            wav_dir = os.path.join(self.data_dir, "wav")
            return {
                os.path.splitext(filename)[0]: os.path.join(wav_dir, filename)
                for filename in sorted(os.listdir(wav_dir))
            }

    def _load_wav(self, wav_rxfilename, start=0, end=None):
        """This function reads audio file and return data in numpy.float32 array.
        "lru_cache" holds recently loaded audio so that can be called
        many times on the same audio file.
        OPTIMIZE: controls lru_cache size for random access,
        considering memory size
        """
        if wav_rxfilename.endswith("|"):
            # input piped command
            p = subprocess.Popen(
                wav_rxfilename[:-1],
                shell=True,
                stdout=subprocess.PIPE,
            )
            data, samplerate = sf.read(
                io.BytesIO(p.stdout.read()),
                dtype="float32",
            )
            # cannot seek
            data = data[start:end]
        elif wav_rxfilename == "-":
            # stdin
            data, samplerate = sf.read(sys.stdin, dtype="float32")
            # cannot seek
            data = data[start:end]
        else:
            # normal wav file
            data, samplerate = sf.read(wav_rxfilename, start=start, stop=end)
        return data, samplerate

    def _load_utt2spk(self, utt2spk_file):
        """Returns dictionary { uttid: spkid }."""
        lines = [line.strip().split(None, 1) for line in open(utt2spk_file)]
        return {x[0]: x[1] for x in lines}

    def _load_spk2utt(self, spk2utt_file):
        """Returns dictionary { spkid: list of uttids }."""
        if not os.path.exists(spk2utt_file):
            return None
        lines = [line.strip().split() for line in open(spk2utt_file)]
        return {x[0]: x[1:] for x in lines}

    def _load_reco2dur(self, reco2dur_file):
        """Returns dictionary { recid: duration }."""
        if not os.path.exists(reco2dur_file):
            return None
        lines = [line.strip().split(None, 1) for line in open(reco2dur_file)]
        return {x[0]: float(x[1]) for x in lines}

    def _process_wav(self, wav_rxfilename, process):
        """This function returns preprocessed wav_rxfilename.
        Args:
            wav_rxfilename:
                input
            process:
                command which can be connected via pipe, use stdin and stdout
        Returns:
            wav_rxfilename: output piped command
        """
        if wav_rxfilename.endswith("|"):
            # input piped command
            return wav_rxfilename + process + "|"
        # stdin "-" or normal file
        return "cat {0} | {1} |".format(wav_rxfilename, process)

    def _extract_segments(self, wavs, segments=None):
        """This function returns generator of segmented audio.
        Yields (utterance id, numpy.float32 array).
        TODO?: sampling rate is not converted.
        """
        if segments is not None:
            # segments should be sorted by rec-id
            for seg in segments:
                wav = wavs[seg["rec"]]
                data, samplerate = self.load_wav(wav)
                st_sample = np.rint(seg["st"] * samplerate).astype(int)
                et_sample = np.rint(seg["et"] * samplerate).astype(int)
                yield seg["utt"], data[st_sample:et_sample]
        else:
            # segments file not found,
            # wav.scp is used as segmented audio list
            for rec in wavs:
                data, samplerate = self.load_wav(wavs[rec])
                yield rec, data



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ expert.py ]
#   Synopsis     [ the phone linear downstream wrapper ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import os
import math
import torch
import random
import time
#-------------#
import torch
import torch.nn as nn
import numpy as np
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence
#-------------#
from .model import Model
from .dataset import AudioBatchData, SpeakerVerifi_test, SpeakerVerifi_dev
from .model import GE2E
from .utils import EER, compute_metrics


class DownstreamExpert(nn.Module):
    """
    Used to handle downstream-specific operations
    eg. downstream forward, metric computation, contents to log

    Note 1.
        dataloaders should output in the following format:

        [[wav1, wav2, ...], your_other_contents, ...]

        where wav1, wav2 ... are in variable length
        and wav1 is in torch.FloatTensor
    """

    def __init__(self, upstream_dim, downstream_expert, **kwargs):
        super(DownstreamExpert, self).__init__()
        self.upstream_dim = upstream_dim
        self.downstream = downstream_expert
        self.datarc = downstream_expert['datarc']
        self.modelrc = downstream_expert['modelrc']
        self.seed = kwargs['seed']

        self.train_dataset = AudioBatchData(**self.datarc['train'],batch_size=self.datarc['train_batch_size'])
        self.dev_dataset = SpeakerVerifi_dev(**self.datarc['dev'])
        self.test_dataset = SpeakerVerifi_test(**self.datarc['test'])
        self.connector = nn.Linear(upstream_dim,self.modelrc['input_dim'])

        self.model = Model(input_dim=self.modelrc['input_dim'], agg_module=self.modelrc['agg_module'],  config=self.modelrc)
        self.objective = GE2E()
        self.score_fn  = nn.CosineSimilarity(dim=-1)
        self.eval_metric = EER

    def _get_train_dataloader(self, dataset):
        return self.train_dataset.getDataLoader(batchSize=1, numWorkers=0)

    def _get_eval_dataloader(self, dataset):
        return DataLoader(
            dataset, batch_size=self.datarc['eval_batch_size'],
            shuffle=False, num_workers=self.datarc['num_workers'],
            collate_fn=dataset.collate_fn
        )

    # Interface
    def get_train_dataloader(self):
        return self._get_train_dataloader(self.train_dataset)

    # Interface
    def get_dev_dataloader(self):
        return self._get_eval_dataloader(self.dev_dataset)

    # Interface
    def get_test_dataloader(self):
        return self._get_eval_dataloader(self.test_dataset)

    # Interface
    def forward(self, features, lengths, labels,
                records=None, logger=None, prefix=None, global_step=0, **kwargs):
        """
        Args:
            features:
                the features extracted by upstream
                put in the device assigned by command-line args

            labels:
                the frame-wise phone labels

            records:
                defaultdict(list), by appending scalars into records,
                these scalars will be averaged and logged on Tensorboard

            logger:
                Tensorboard SummaryWriter, given here for logging/debugging
                convenience, please use "self.downstream/your_content_name" as key
                name to log your customized contents

            global_step:
                global_step in runner, which is helpful for Tensorboard logging

        Return:
            loss:
                the loss to be optimized, should not be detached
        """
        features_pad = pad_sequence(features, batch_first=True)
        
        attention_mask = [torch.ones((feature.shape[0])) for feature in features] 

        attention_mask_pad = pad_sequence(attention_mask,batch_first=True)

        attention_mask_pad = (1.0 - attention_mask_pad) * -100000.0

        features_pad = self.connector(features_pad)
        agg_vec = self.model(features_pad, attention_mask_pad.cuda())

        # normalize to unit vector 
        agg_vec = agg_vec / torch.norm(agg_vec, dim=-1).unsqueeze(-1)

        if self.training:
            
            GE2E_matrix = agg_vec.reshape(-1, self.train_dataset.utter_number, agg_vec.shape[-1])
            loss = self.objective(GE2E_matrix)
            
            return loss
        
        else:
            vec1, vec2 = self.separate_data(agg_vec, labels)
            scores = self.score_fn(vec1,vec2).squeeze().cpu().detach().tolist()
            ylabels = torch.stack(labels).cpu().detach().long().tolist()

            if len(ylabels) > 1:
                records['scores'].extend(scores)
                records['ylabels'].extend(ylabels)
            else:
                records['scores'].append(scores)
                records['ylabels'].append(ylabels)

            return torch.tensor(0)

        # interface
    def log_records(self, records, logger, prefix, global_step, **kwargs):
        """
        Args:
            records:
                defaultdict(list), contents already appended

            logger:
                Tensorboard SummaryWriter
                please use f'{prefix}your_content_name' as key name
                to log your customized contents

            prefix:
                used to indicate downstream and train/test on Tensorboard
                eg. 'phone/train-'

            global_step:
                global_step in runner, which is helpful for Tensorboard logging
        """

        if not self.training:

            EER_result =self.eval_metric(np.array(records['ylabels']), np.array(records['scores']))

            records['EER'] = EER_result[0]

            logger.add_scalar(
                f'{prefix}'+'EER',
                records['EER'],
                global_step=global_step
            )
        
    def separate_data(self, agg_vec, ylabel):

        total_num = len(ylabel) 
        feature1 = agg_vec[:total_num]
        feature2 = agg_vec[total_num:]
        
        return feature1, feature2


# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/npc/expert.py ]
#   Synopsis     [ the npc wrapper ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import torch
from torch.nn.utils.rnn import pad_sequence

from ..interfaces import UpstreamBase
from .npc import NPC
from .audio import create_transform


class UpstreamExpert(UpstreamBase):
    def __init__(self, ckpt, **kwargs):
        super().__init__(**kwargs)

        ckpt = torch.load(ckpt, map_location="cpu")
        config = ckpt["config"]
        self.preprocessor, feat_dim = create_transform(config["data"]["audio"])
        self.model = NPC(feat_dim, **config["model"]["paras"])
        self.model.load_state_dict(ckpt["model"])

        if len(self.hooks) == 0:
            for block_id, _ in enumerate(self.model.blocks):
                self.add_hook(
                    f"self.model.blocks[{block_id}]",
                    lambda input, output: output.transpose(1, 2),
                )

            for masked_conv_id, _ in enumerate(self.model.masked_convs):
                self.add_hook(
                    f"self.model.masked_convs[{masked_conv_id}]",
                    lambda input, output: output,
                )

            self.add_hook("self.model", lambda input, output: output[1])

    def get_downsample_rates(self, key: str) -> int:
        return 160

    def forward(self, wavs):
        features = [self.preprocessor(wav.unsqueeze(0)) for wav in wavs]
        features = pad_sequence(features, batch_first=True)

        predicted_BxLxM, features = self.model(features, testing=not self.training)

        # This forward function only does the model forward
        # The return dict is then handled by UpstreamBase's hooks



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/apc/expert.py ]
#   Synopsis     [ the apc wrapper ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import torch

import torch
from torch.nn.utils.rnn import pad_sequence, pad_packed_sequence

from ..interfaces import UpstreamBase
from .apc import APC
from .audio import create_transform


class UpstreamExpert(UpstreamBase):
    def __init__(self, ckpt, **kwargs):
        super().__init__(**kwargs)

        ckpt = torch.load(ckpt, map_location="cpu")
        config = ckpt["config"]

        self.preprocessor, feat_dim = create_transform(config["data"]["audio"])
        self.model = APC(feat_dim, **config["model"]["paras"])
        self.model.load_state_dict(ckpt["model"])

        if len(self.hooks) == 0:
            self.add_hook(
                "self.model.rnn_layers[1]",
                lambda input, output: pad_packed_sequence(input[0], batch_first=True)[
                    0
                ],
            )
            self.add_hook(
                "self.model.rnn_layers[2]",
                lambda input, output: pad_packed_sequence(input[0], batch_first=True)[
                    0
                ],
            )
            self.add_hook("self.model", lambda input, output: output[1])

    def get_downsample_rates(self, key: str) -> int:
        return 160

    def forward(self, wavs):
        features = [self.preprocessor(wav.unsqueeze(0)) for wav in wavs]
        feat_lengths = [len(feat) for feat in features]

        features = pad_sequence(features, batch_first=True)
        feat_lengths = torch.LongTensor(feat_lengths)

        predicted_BxLxM, features = self.model(
            features, feat_lengths, testing=not self.training
        )

        # This forward function only does the model forward
        # The return dict is then handled by UpstreamBase's hooks



import os
import math
import torch
import random

import torch.nn as nn
import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader, DistributedSampler
from torch.distributed import is_initialized
from torch.nn.utils.rnn import pad_sequence

import sentencepiece
import sacrebleu
from tqdm.auto import tqdm
import editdistance
import fairseq
from argparse import Namespace
from .S3prl_SpeechToTextTask import S3prl_SpeechToTextTask
from .AdditionalDataset import AdditionalDataset
from fairseq.models.speech_to_text.s2t_transformer import TransformerDecoderScriptable, S2TTransformerModel
from fairseq.models.transformer import Embedding
from fairseq.data import Dictionary, encoders
import string

class DownstreamExpert(nn.Module):
    """
    Used to handle downstream-specific operations
    eg. downstream forward, metric computation, contents to log
    """

    def __init__(self, upstream_dim, upstream_rate, downstream_expert, expdir, **kwargs):
        """
        Args:
            upstream_dim: int
                Different upstream will give different representation dimension
                You might want to first project them to the same dimension

            upstream_rate: int
                160: for upstream with 10 ms per frame
                320: for upstream with 20 ms per frame
            
            downstream_expert: dict
                The 'downstream_expert' field specified in your downstream config file
                eg. downstream/example/config.yaml

            expdir: string
                The expdir from command-line argument, you should save all results into
                this directory, like some logging files.

            **kwargs: dict
                All the arguments specified by the argparser in run_downstream.py
                and all the other fields in config.yaml, in case you need it.
                
                Note1. Feel free to add new argument for __init__ as long as it is
                a command-line argument or a config field. You can check the constructor
                code in downstream/runner.py
        """

        super(DownstreamExpert, self).__init__()

        print(downstream_expert)

        self.expdir = expdir
        self.src_lang = downstream_expert['src_lang']
        self.tgt_lang = downstream_expert['tgt_lang']
        self.post_process = downstream_expert['post_process']
        self.output_prefix = downstream_expert['output_prefix']
        self.upstream_rate = upstream_rate

        self.datarc = downstream_expert['datarc']
        self.max_positions = downstream_expert['modelrc']['max_source_positions']

        self.upstream_rate = downstream_expert.get('upstream_rate', upstream_rate)
        if self.upstream_rate < 0:
            self.upstream_rate = upstream_rate
        assert self.upstream_rate % upstream_rate == 0
        self.downsample_ratio = int(self.upstream_rate / upstream_rate)
        self.downsample_method = downstream_expert.get('downsample_method', 'drop')
        if self.downsample_method == 'concat':
            upstream_dim *= self.downsample_ratio

        self.task = S3prl_SpeechToTextTask.setup_task(Namespace(**downstream_expert['taskrc']))
        self.task.upstream_rate = self.upstream_rate

        self.data_dir = downstream_expert['taskrc']['data']

        self.criterion = self.task.build_criterion(Namespace(**downstream_expert['criterionrc']))

        modelrc = Namespace(**downstream_expert['modelrc'])
        assert modelrc.arch in fairseq.models.ARCH_CONFIG_REGISTRY
        fairseq.models.ARCH_CONFIG_REGISTRY[modelrc.arch](modelrc)
        self.model = self.task.build_model(modelrc, upstream_dim)

        self.generator = self.task.build_generator([self.model], Namespace(**downstream_expert['generatorrc']))
        self.batch_itr = {}
        
        self.use_asr = downstream_expert['taskrc']['use_asr']

        if self.use_asr:

            rc = downstream_expert['asrrc']
            self.asr_datarc = rc['datarc']
            self.asr_weight = rc['weight']
            self.asr_dict = Dictionary.load(f"{self.data_dir}/{rc['vocab_file']}")
            
            asr_bperc = rc['bpe_tokenizer'].copy()
            asr_bperc['sentencepiece_model'] = f"{self.data_dir}/{asr_bperc['sentencepiece_model']}" 
            self.asr_bpe = encoders.build_bpe(Namespace(**asr_bperc))
            
            self.asr_task = S3prl_SpeechToTextTask.setup_task(Namespace(**downstream_expert['taskrc']))
            self.asr_dict.add_symbol('<blank>')
            self.asr_task.tgt_dict = self.asr_dict
            self.asr_head = nn.Linear(modelrc.encoder_embed_dim, len(self.asr_dict))
            self.additional_dataset = {}

        self.register_buffer('best_score', torch.zeros(1))

    # Interface
    def get_dataloader(self, split, epoch: int = 0):
        """
        Args:
            mode: string
                'train', 'dev' or 'test'

        Return:
            a torch.utils.data.DataLoader returning each batch in the format of:

            [wav1, wav2, ...], your_other_contents1, your_other_contents2, ...

            where wav1, wav2 ... are in variable length
            each wav is torch.FloatTensor in cpu with:
                1. dim() == 1
                2. sample_rate == 16000
                3. directly loaded by torchaudio
        """


        data_split = self.datarc[split]

        # load dataset
        if data_split not in self.batch_itr:
            
            # dataset will truncate the input wav according to model's max_position
            self.task.load_dataset(split=data_split, max_feature_len=self.max_positions)
            
            # it must not have invalid_inputs due to truncation
            self.batch_itr[data_split] = self.task.get_batch_iterator(
                self.task.dataset(data_split),
                max_tokens=self.datarc['max_tokens'],
                max_positions=self.max_positions,
                num_workers=self.datarc['num_workers'],
                ignore_invalid_inputs = False,
                epoch=epoch+1,
            )

        # # fairseq's dataloader
        # # note: refreshing is needed for each epoch
        return self.batch_itr[data_split].next_epoch_itr()

    # Interface
    def forward(self, mode, features, input_dict, records, **kwargs):
        """
        Args:
            mode: string
                'train', 'dev' or 'test' for this forward step

            features:
                list of unpadded features [feat1, feat2, ...]
                each feat is in torch.FloatTensor and already
                put in the device assigned by command-line args

            your_other_contents1, ... :
                in the order defined by your dataloader (dataset + collate_fn)
                these are all in cpu, and you can move them to the same device
                as features

            records:
                defaultdict(list), by appending contents into records,
                these contents can be averaged and logged on Tensorboard
                later by self.log_records (also customized by you)

                Note1. downstream/runner.py will call self.log_records
                    1. every `log_step` during training
                    2. once after evalute the whole dev/test dataloader

                Note2. `log_step` is defined in your downstream config
                eg. downstream/example/config.yaml

        Return:
            loss:
                the loss to be optimized, should not be detached
                a single scalar in torch.FloatTensor
        """
        device = features[0].device
        features = self.downsample(features)
        features_length = torch.LongTensor([len(feature) for feature in features])
        features = pad_sequence(features, batch_first=True, padding_value=0.0)

        input_dict['net_input']['src_tokens'] = features
        input_dict['net_input']['src_lengths'] = features_length

        input_dict = fairseq.utils.move_to_cuda(input_dict, device=device)

        if self.use_asr:
            asr_input_dict = self._create_asr_input_dict(input_dict, mode)
            asr_input_dict = fairseq.utils.move_to_cuda(asr_input_dict, device=device)

        loss = torch.FloatTensor(0)

        if mode in ['train', 'dev']:

            encoder_out = self.model.encoder(
                src_tokens=input_dict['net_input']['src_tokens'], src_lengths=input_dict['net_input']['src_lengths']
            )

            st_decoder_out = self.model.decoder(
                prev_output_tokens=input_dict['net_input']['prev_output_tokens'],
                encoder_out=encoder_out
            )

            st_loss, _ = self.criterion.compute_loss(
                self.model,
                st_decoder_out,
                input_dict,
            )

            loss = st_loss

            if self.use_asr:

                asr_loss = self.count_asr_loss(encoder_out, asr_input_dict)
                loss = (1-self.asr_weight) * st_loss + self.asr_weight * asr_loss

            loss /= input_dict['nsentences']

            records['loss'].append(loss.item())

            if self.use_asr:

                records['st_loss'].append(st_loss.item())
                records['asr_loss'].append(asr_loss.item())


        if mode in ['dev', 'test']:

            records['ids'] += input_dict['id'].cpu().tolist()
            records['utt_ids'] += input_dict['utt_id']

            hyps, refs = self._inference_step(input_dict)
            records['hyps'] += hyps
            records['refs'] += refs

            if self.use_asr:
                asr_hyps, asr_refs = self._inference_step_asr(asr_input_dict)
                records['asr_hyps'] += asr_hyps
                records['asr_refs'] += asr_refs

        return loss

    def downsample(self, features):

        if self.downsample_ratio == 1:
            return features
        
        new_features = []
        
        for feature in features:
            if self.downsample_method == 'drop':

                feature = feature[::self.downsample_ratio]
            
            elif self.downsample_method == 'concat':
                
                N = feature.size(0) % self.downsample_ratio
                if N != 0:
                    feature = F.pad(feature, (0, 0, 0, self.downsample_ratio-N))
                feature = feature.view(feature.size(0)//self.downsample_ratio, feature.size(1)*self.downsample_ratio)
            
            elif self.downsample_method == 'average':

                N = feature.size(0) % self.downsample_ratio
                if N != 0:
                    feature = F.pad(feature, (0, 0, 0, self.downsample_ratio-N))
                feature = feature.view(feature.size(0)//self.downsample_ratio, self.downsample_ratio, feature.size(1)).mean(dim=1)

            else:
                raise NotImplementedError
            
            new_features.append(feature)
        
        return new_features

    def _create_asr_input_dict(self, input_dict, mode):

        if mode not in self.additional_dataset:
            
            dataset = AdditionalDataset.from_tsv(
                f'{self.data_dir}/{self.datarc[mode]}.tsv',
                self.asr_datarc['key'],
                self.asr_dict,
                self.asr_bpe,
            )
            self.additional_dataset[mode] = dataset

        additional_data = self.additional_dataset[mode].get_addtional_input(input_dict['id'])

        asr_input_dict = input_dict.copy()
        asr_input_dict['net_input'] = input_dict['net_input'].copy()
        asr_input_dict['net_input']['prev_output_tokens'] = additional_data['prev_output_tokens']
        asr_input_dict['target'] = additional_data['target']
        asr_input_dict['target_lengths'] = additional_data['target_lengths']
        asr_input_dict['ntokens'] = additional_data['ntokens']

        return asr_input_dict

    def count_asr_loss(self, encoder_out, input_dict):

        hidden = encoder_out['encoder_out'][0] # T x B x C
        log_prob = self.asr_head(hidden).log_softmax(2)

        hidden_length = self.model.encoder.subsample.get_out_seq_lens_tensor(input_dict['net_input']['src_lengths'])
        
        targets = input_dict['target']
        target_lengths = input_dict['target_lengths']

        loss = nn.functional.ctc_loss(log_prob, targets, hidden_length, target_lengths, blank=self.asr_dict.index('<blank>'), reduction='sum', zero_infinity=True)

        return loss

    def _decode(self, toks, dictionary):

        toks = toks[toks != dictionary.pad()]

        s = dictionary.string(
            toks.int().cpu(),
            self.post_process,
        )

        return s if s else "<unk>"

    def _inference_step(self, input_dict):
        output = self.generator.generate([self.model], input_dict)

        hyps = []
        refs = []

        for i in range(len(output)):

            hyps.append(
                self._decode(output[i][0]["tokens"], self.task.target_dictionary)
            )

            refs.append(
                self._decode(input_dict['target'][i], self.task.target_dictionary)
            )

        return hyps, refs

    def _inference_step_asr(self, input_dict):
        
        encoder_out = self.model.encoder(
            src_tokens=input_dict['net_input']['src_tokens'], src_lengths=input_dict['net_input']['src_lengths']
        )

        hidden = encoder_out['encoder_out'][0] # TxBxC
        logit = self.asr_head(hidden)

        predict = logit.argmax(dim=-1).transpose(0, 1)
        
        hyps = []
        refs = []

        for i in range(len(predict)):

            predict_ids = predict[i].unique_consecutive()
            predict_ids = predict_ids[predict_ids != self.asr_dict.index('<blank>')]

            hyps.append(
                self._decode(predict_ids, self.asr_dict)
            )

            refs.append(
                self._decode(input_dict['target'][i], self.asr_dict)
            )
        
        return hyps, refs

    def _metric(self, hyps, refs):

        tok = 'zh' if self.tgt_lang == 'zh' else '13a'
        bleu = sacrebleu.corpus_bleu(hyps, [refs], tokenize=tok)

        return bleu

    def _asr_metric(self, hyps, refs):

        ce = 0
        we = 0
        c_total = 0
        w_total = 0

        for hyp, ref in zip(hyps, refs):

            normalized_hyp = hyp.translate(str.maketrans('', '', "".join(list(set(string.punctuation)-set("'-"))))).lower()
            normalized_ref = ref.translate(str.maketrans('', '', "".join(list(set(string.punctuation)-set("'-"))))).lower()

            ce += editdistance.eval(normalized_hyp, normalized_ref)
            c_total += len(normalized_ref)

            hyp_w = normalized_hyp.split()
            ref_w = normalized_ref.split()

            we += editdistance.eval(hyp_w, ref_w)
            w_total += len(ref_w)


        cer = ce / c_total
        wer = we / w_total

        return cer, wer

    # interface
    def log_records(self, mode, records, logger, global_step, batch_ids, total_batch_num, **kwargs):
        """
        Args:
            mode: string
                'train':
                    records and batchids contain contents for `log_step` batches
                    `log_step` is defined in your downstream config
                    eg. downstream/example/config.yaml

                'dev' or 'test' :
                    records and batchids contain contents for the entire evaluation dataset

            records:
                defaultdict(list), contents already prepared by self.forward

            logger:
                Tensorboard SummaryWriter
                please use f'{your_task_name}/{mode}-{key}' as key name to log your contents,
                preventing conflict with the logging of other tasks

            global_step:
                The global_step when training, which is helpful for Tensorboard logging

            batch_ids:
                The batches contained in records when enumerating over the dataloader

            total_batch_num:
                The total amount of batches in the dataloader
        
        Return:
            a list of string
                Each string is a filename we wish to use to save the current model
                according to the evaluation result, like the best.ckpt on the dev set
                You can return nothing or an empty list when no need to save the checkpoint
        """
        save_names = []

        if mode in ['train', 'dev']:

            ave_loss = sum(records['loss'])/len(records['loss'])
            logger.add_scalar(
                f'st/{mode}-loss',
                ave_loss,
                global_step=global_step
            )

            if self.use_asr:

                ave_st_loss = sum(records['st_loss'])/len(records['st_loss'])
                logger.add_scalar(
                    f'st/{mode}-st_loss',
                    ave_st_loss,
                    global_step=global_step
                )

                ave_asr_loss = sum(records['asr_loss'])/len(records['asr_loss'])
                logger.add_scalar(
                    f'st/{mode}-asr_loss',
                    ave_asr_loss,
                    global_step=global_step
                )

        if mode in ['dev', 'test']:

            bleu = self._metric(records['hyps'], records['refs'])
            logger.add_scalar(
                f'st/{mode}-bleu',
                bleu.score,
                global_step=global_step
            )
            for i in range(4):
                logger.add_scalar(
                    f'st/{mode}-bleu{i+1}',
                    bleu.precisions[i],
                    global_step=global_step
                )

            if bleu.score > self.best_score and mode == 'dev':
                self.best_score = torch.ones(1) * bleu.score
                save_names.append(f'{mode}-best.ckpt') 
            
            with open(f'{self.expdir}/{self.output_prefix}-st-{mode}.tsv', 'w') as f:
                print('utt_id', 'hyp', 'ref', sep='\t', file=f)
                results = list(zip(records['ids'], records['hyps'], records['refs'], records['utt_ids']))
                results.sort(key=lambda x: x[0])
                for idx, hyp, ref, utt_id in results:
                    print(utt_id, hyp, ref, sep='\t', file=f)

            print(bleu)

            if self.use_asr:

                cer, wer = self._asr_metric(records['asr_hyps'], records['asr_refs'])
                logger.add_scalar(
                    f'st/{mode}-asr-cer',
                    cer,
                    global_step=global_step
                )
                logger.add_scalar(
                    f'st/{mode}-asr-wer',
                    wer,
                    global_step=global_step
                )

                with open(f'{self.expdir}/{self.output_prefix}-asr-{mode}.tsv', 'w') as f:
                    print('utt_id', 'hyp', 'ref', sep='\t', file=f)
                    results = list(zip(records['ids'], records['asr_hyps'], records['asr_refs'], records['utt_ids']))
                    results.sort(key=lambda x: x[0])
                    for idx, hyp, ref, utt_id in results:
                        print(utt_id, hyp, ref, sep='\t', file=f)

                tqdm.write(f'[cer]:{cer}, [wer]:{wer}')
        
        return save_names


import pytest
import shutil
import logging
import tempfile
import traceback
from pathlib import Path
from subprocess import check_call

import torch
from s3prl.nn import S3PRLUpstream, Featurizer
from s3prl.util.pseudo_data import get_pseudo_wavs
from s3prl.util.download import _urls_to_filepaths

logger = logging.getLogger(__name__)

TEST_MORE_ITER = 2
TRAIN_MORE_ITER = 5
SAMPLE_RATE = 16000
EXTRACTED_GT_DIR = Path(__file__).parent.parent / "sample_hidden_states"

# Expect the following directory structure:
#
# -- s3prl/  (repository root)
# ---- s3prl/  (package root)
# ---- test/
# ------- test_upstream.py
# ---- sample_hidden_states/


def _prepare_sample_hidden_states():
    if not EXTRACTED_GT_DIR.is_dir():
        with tempfile.TemporaryDirectory() as tempdir:
            tempdir = Path(tempdir)
            tempdir.mkdir(exist_ok=True, parents=True)

            logger.info("Downloading extracted sample hidden states...")
            check_call("git lfs install".split(), cwd=tempdir)
            check_call(
                "git clone https://huggingface.co/datasets/s3prl/sample_hidden_states".split(),
                cwd=tempdir,
            )
            shutil.move(
                str(tempdir / "sample_hidden_states"), str(EXTRACTED_GT_DIR.parent)
            )
    else:
        logger.info(f"{EXTRACTED_GT_DIR} exists. Perform git pull...")
        check_call("git pull".split(), cwd=EXTRACTED_GT_DIR)


def _extract_feat(model: S3PRLUpstream, seed: int = 0):
    wavs, wavs_len = get_pseudo_wavs(seed=seed, padded=True)
    all_hs, all_lens = model(wavs, wavs_len)
    return all_hs


def _all_hidden_states_same(hs1, hs2):
    for h1, h2 in zip(hs1, hs2):
        if h1.size(1) != h2.size(1):
            min_seqlen = min(h1.size(1), h2.size(1))
            h1 = h1[:, :min_seqlen, :]
            h2 = h2[:, :min_seqlen, :]
        assert torch.allclose(h1, h2)


def _load_ground_truth(name: str):
    source = f"{EXTRACTED_GT_DIR}/{name}.pt"
    if source.startswith("http"):
        path = _urls_to_filepaths(source)
    else:
        path = source
    return torch.load(path)


def _compare_with_extracted(name: str):
    model = S3PRLUpstream(name)
    model.eval()

    with torch.no_grad():
        hs = _extract_feat(model)
        hs_gt = _load_ground_truth(name)

        _all_hidden_states_same(hs, hs_gt)

        for i in range(TEST_MORE_ITER):
            more_hs = _extract_feat(model)
            for h1, h2 in zip(hs, more_hs):
                assert torch.allclose(
                    h1, h2
                ), "should have deterministic representation in eval mode"

        for i in range(TEST_MORE_ITER):
            more_hs = _extract_feat(model, seed=i + 1)
            assert len(hs) == len(
                more_hs
            ), "should have deterministic num_layer in eval mode"

    model.train()
    for i in range(TRAIN_MORE_ITER):
        more_hs = _extract_feat(model, seed=i + 1)
        assert len(hs) == len(
            more_hs
        ), "should have deterministic num_layer in train mode"


def _test_model(name: str):
    """
    Test the upstream with the name: 'name' can successfully forward and backward
    """
    with torch.autograd.set_detect_anomaly(True):
        model = S3PRLUpstream(name)
        hs = _extract_feat(model)
        h_sum = 0
        for h in hs:
            h_sum = h_sum + h.sum()
        h_sum.backward()


"""
Test cases ensure that all upstreams are working and are same with pre-extracted features
"""


@pytest.mark.slow
def test_all_model():
    _prepare_sample_hidden_states()

    options = [
        name
        for name in S3PRLUpstream.available_names(only_registered_ckpt=True)
        if (not name == "customized_upstream")
        and (
            not "mos" in name
        )  # mos models do not have hidden_states key. They only return a single mos score
        and (
            not "stft_mag" in name
        )  # stft_mag upstream must past the config file currently and is not so important. So, skip the test now
        and (
            not "pase" in name
        )  # pase_plus needs lots of dependencies and is difficult to be tested and is not very worthy today
        and (
            not name == "xls_r_1b"
        )  # skip due to too large model, too long download time
        and (
            not name == "xls_r_2b"
        )  # skip due to too large model, too long download time
    ]

    tracebacks = []
    for name in options:
        logger.info(f"Testing upstream: '{name}'")
        try:
            _compare_with_extracted(name)
            _test_model(name)

        except Exception as e:
            logger.error(f"{name}\n{traceback.format_exc()}")
            tb = traceback.format_exc()
            tracebacks.append((name, tb))

    if len(tracebacks) > 0:
        for name, tb in tracebacks:
            logger.error(f"Error in {name}:\n{tb}")
        logger.error(f"All failed models:\n{[name for name, _ in tracebacks]}")
        assert False


def test_one_model(upstream_name: str):
    if upstream_name is None:
        return

    _prepare_sample_hidden_states()
    _compare_with_extracted(upstream_name)
    _test_model(upstream_name)


@pytest.mark.parametrize("layer_selections", [None, [0, 4, 9]])
@pytest.mark.parametrize("normalize", [False, True])
def test_featurizer(layer_selections, normalize):
    model = S3PRLUpstream("hubert")
    featurizer = Featurizer(
        model, layer_selections=layer_selections, normalize=normalize
    )

    wavs, wavs_len = get_pseudo_wavs(padded=True)
    all_hs, all_lens = model(wavs, wavs_len)
    hs, hs_len = featurizer(all_hs, all_lens)

    assert isinstance(hs, torch.FloatTensor)
    assert isinstance(hs_len, torch.LongTensor)


def test_upstream_properties():
    model = S3PRLUpstream("hubert")
    featurizer = Featurizer(model)
    assert isinstance(model.hidden_sizes, (tuple, list)) and isinstance(
        model.hidden_sizes[0], int
    )
    assert isinstance(model.downsample_rates, (tuple, list)) and isinstance(
        model.downsample_rates[0], int
    )
    assert isinstance(featurizer.output_size, int)
    assert isinstance(featurizer.downsample_rate, int)



import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--src", required=True)
parser.add_argument("--tgt", required=True)
args = parser.parse_args()

def read_file(path, callback=lambda x: x, sep=" ", default_value=""):
    content = {}
    with open(path, "r") as file:
        lines = file.readlines()
        for idx, line in enumerate(lines):
            fields = line.strip().split(sep, maxsplit=1)
            if len(fields) > 1:
                filename, value = fields
            else:
                filename = fields[0]
                value = default_value
            content[filename] = callback(value)
    return content

src = read_file(args.src)
tgt = read_file(args.tgt)

match = []
for key in src.keys():
    truth = tgt[key]
    predict = src[key]
    if truth == predict:
        match.append(1)
    else:
        match.append(0)

print(sum(match) / len(match))


# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ utility/fix_ckpt.py ]
#   Synopsis     [ scripts to fix older checkpoints ]
#   Author       [ Andy T. Liu (Andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


"""
Usage:
This .py helps fix the `torch serialization ModuleNotFoundError` issue, 
which occurs when the model.py directory is changed.
Make sure you understand exactly what this script does before proceeding.
"""


###############
# IMPORTATION #
###############
import os
import sys
import torch


def check_model_equiv(model1, model2):
    for p1, p2 in zip(model1.parameters(), model2.parameters()):
        if p1.data.ne(p2.data).sum() > 0:
            return False
        if not torch.equal(p1[0], p2[0]):
            return False
        if not torch.equal(p1[1].data, p2[1].data):
            return False
    return True


def copyParams(module_src, module_dest):
    params1 = module_src.named_parameters()
    params2 = module_dest.named_parameters()
    dict_params2 = dict(params2)
    for name1, param1 in params1:
        if name1 in dict_params2:
            dict_params2[name1].data.copy_(param1.data)


def main():

    input_ckpt = sys.argv[1]

    # load model with old setting
    from transformer.nn_transformer import SPEC_TRANSFORMER
    options = {'ckpt_file'     : input_ckpt,
                'load_pretrain' : 'True',
                'no_grad'       : 'True',
                'dropout'       : 'default',
                'spec_aug'      : 'False',
                'spec_aug_prev' : 'True',
                'weighted_sum'  : 'False',
                'select_layer'  : -1,
                'permute_input' : 'False' }
    old_transformer = SPEC_TRANSFORMER(options, inp_dim=-1)

    # build model with new setting
    from s3prl.upstream.mockingjay.model import TransformerForMaskedAcousticModel
    model = TransformerForMaskedAcousticModel(old_transformer.model_config, old_transformer.inp_dim, old_transformer.inp_dim).to(torch.device('cuda'))

    # load old to new
    assert not check_model_equiv(old_transformer.model, model.Transformer)
    copyParams(old_transformer.model, model.Transformer)
    assert check_model_equiv(old_transformer.model, model.Transformer)

    assert not check_model_equiv(old_transformer.SpecHead, model.SpecHead)
    copyParams(old_transformer.SpecHead, model.SpecHead)
    assert check_model_equiv(old_transformer.SpecHead, model.SpecHead)

    global_step = old_transformer.all_states['Global_step']
    settings = old_transformer.all_states['Settings']

    # save
    all_states = {
        'SpecHead': model.SpecHead.state_dict(),
        'Transformer': model.Transformer.state_dict(),
        'Global_step': global_step,
        'Settings': settings
    }
    new_ckpt_path = input_ckpt.replace('.ckpt', '-new.ckpt')
    torch.save(all_states, new_ckpt_path)

    print('Done fixing ckpt: ', input_ckpt, 'to: ', new_ckpt_path)


if __name__ == '__main__':
    main()



import torch
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence

from ..interfaces import UpstreamBase
from .convert import load_converted_model


class UpstreamExpert(UpstreamBase):
    def __init__(self, ckpt, **kwargs):
        super().__init__(**kwargs)
        model, task_cfg = load_converted_model(ckpt)
        self.model = model
        self.wav_normalize = task_cfg.normalize

        self.model.feature_grad_mult = 0.0
        self.model.encoder.layerdrop = 0.0

        if len(self.hooks) == 0:
            module_name = "self.model.encoder.layers"
            for module_id in range(len(eval(module_name))):
                self.add_hook(
                    f"{module_name}[{module_id}]",
                    lambda input, output: input[0].transpose(0, 1),
                )
            self.add_hook("self.model.encoder", lambda input, output: output[0])

            def postprocess(xs):
                names, hiddens = zip(*xs)
                unpad_len = min([hidden.size(1) for hidden in hiddens])
                hiddens = [hidden[:, :unpad_len, :] for hidden in hiddens]
                return list(zip(names, hiddens))

            self.hook_postprocess = postprocess

    def get_downsample_rates(self, key: str) -> int:
        return 320

    def forward(self, wavs):
        device = wavs[0].device
        if self.wav_normalize:
            wavs = [F.layer_norm(wav, wav.shape) for wav in wavs]

        wav_lengths = torch.LongTensor([len(wav) for wav in wavs]).to(device)
        wav_padding_mask = ~torch.lt(
            torch.arange(max(wav_lengths)).unsqueeze(0).to(device),
            wav_lengths.unsqueeze(1),
        )
        padded_wav = pad_sequence(wavs, batch_first=True)

        results = self.model.extract_features(padded_wav, wav_padding_mask)

        # This forward function only does the model forward
        # The return dict is then handled by UpstreamBase's hooks



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ model.py ]
#   Synopsis     [ Baseline model for speaker diarization ]
#   Author       [ Jiatong ]
#   Copyright    [ Copyright(c), Johns Hopkins University ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import torch
import torch.nn as nn


#########
# MODEL #
#########
class Model(nn.Module):
    def __init__(self, input_dim, output_class_num, rnn_layers, hidden_size, **kwargs):
        super(Model, self).__init__()

        # init attributes
        self.use_rnn = rnn_layers > 0
        if self.use_rnn:
            self.rnn = nn.LSTM(
                input_dim, hidden_size, num_layers=rnn_layers, batch_first=True
            )
            self.linear = nn.Linear(hidden_size, output_class_num)
        else:
            self.linear = nn.Linear(input_dim, output_class_num)

    def forward(self, features):
        features = features.float()
        if self.use_rnn:
            hidden, _ = self.rnn(features)
            predicted = self.linear(hidden)
        else:
            predicted = self.linear(features)
        return predicted



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/baseline/preprocessor.py ]
#   Synopsis     [ the baseline feature preprocessor with torchaudio.transforms as backend ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import torch
import torchaudio
from functools import partial
from torch.nn.utils.rnn import pad_sequence
from torchaudio.transforms import Spectrogram, MelScale, MFCC
from torchaudio.functional import compute_deltas


############
# CONSTANT #
############
N_SAMPLED_PSEUDO_WAV = 2


def get_preprocessor(audio_config, take_first_channel=True, process_input_only=False):
    """
        Args:
            take_first_channel:
                bool
                If True, the preprocessor takes input as: (*, channel=1, waveform_len),
                    where `input` and `target` are taken from the same channel (i.e. the same and single waveform).
                If False, the preprocessor takes input as: (*, channel=2, waveform_len),
                    where `input` and `target` are taken from the specified channel (i.e. multiple correlated views of the same waveform).
                    For example: preprocessor_input = torch.cat([wav_aug1.unsqueeze(1), wav_aug2.unsqueeze(1)], dim=1) # shape: (batch_size, channel=2, seq_len)     
            process_input_only:
                bool
                If True, the preprocessor will process `input`
                If False, the preprocessor will process both `input` and `target`
    """
    assert not (take_first_channel == False and process_input_only == True)

    if not 'target' in audio_config:
        audio_config['target'] = audio_config['input']
        
    input_feat = audio_config['input']
    target_feat = audio_config['target']
    if take_first_channel:
        input_feat['channel'] = 0
        target_feat['channel'] = 0

    preprocessor = OnlinePreprocessor(**audio_config, feat_list=[input_feat, target_feat])
    input_dim, target_dim = [feat.size(-1) for feat in preprocessor()]
    
    if process_input_only:
        del preprocessor
        preprocessor = OnlinePreprocessor(**audio_config, feat_list=[input_feat])
    return preprocessor, input_dim, target_dim


class OnlinePreprocessor(torch.nn.Module):
    def __init__(self, sample_rate=16000, win_ms=25, hop_ms=10, n_freq=201, n_mels=40, n_mfcc=13, feat_list=None, eps=1e-10, **kwargs):
        super(OnlinePreprocessor, self).__init__()
        # save preprocessing arguments
        self._sample_rate = sample_rate
        self._win_ms = win_ms
        self._hop_ms = hop_ms
        self._n_freq = n_freq
        self._n_mels = n_mels
        self._n_mfcc = n_mfcc

        win = round(win_ms * sample_rate / 1000)
        hop = round(hop_ms * sample_rate / 1000)
        n_fft = (n_freq - 1) * 2
        self._win_args = {'n_fft': n_fft, 'hop_length': hop, 'win_length': win}
        self.register_buffer('_window', torch.hann_window(win))
        
        self._stft_args = {'center': True, 'pad_mode': 'reflect', 'normalized': False, 'onesided': True}
        # stft_args: same default values as torchaudio.transforms.Spectrogram & librosa.core.spectrum._spectrogram
        self._stft = partial(torch.stft, **self._win_args, **self._stft_args)
        self._melscale = MelScale(sample_rate=sample_rate, n_mels=n_mels)
        self._mfcc_trans = MFCC(sample_rate=sample_rate, n_mfcc=n_mfcc, log_mels=True, melkwargs=self._win_args)
        self._istft = partial(torch.istft, **self._win_args, **self._stft_args)
        
        self.feat_list = feat_list
        self.register_buffer('_pseudo_wavs', torch.randn(N_SAMPLED_PSEUDO_WAV, sample_rate))
        self.eps = eps

    def _magphase(self, spectrogram):
        # Replace the deprecated private member function self._magphase
        # Ref. https://github.com/pytorch/audio/issues/1337
        # original: `self._magphase = partial(torchaudio.functional.magphase, power=2)`
        spectrogram = torch.view_as_complex(spectrogram)
        return spectrogram.abs().pow(exponent=2), spectrogram.angle()

    def _check_list(self, feat_list):
        if feat_list is None:
            feat_list = self.feat_list
        assert type(feat_list) is list
        return feat_list

    def _transpose_list(self, feats):
        return [feat.transpose(-1, -2).contiguous() if type(feat) is torch.Tensor else feat for feat in feats]

    @classmethod
    def get_feat_config(cls, feat_type, channel=0, log=False, delta=0, cmvn=False):
        assert feat_type in ['wav', 'complx', 'linear', 'phase', 'mel', 'mfcc']
        assert type(channel) is int
        assert type(log) is bool
        assert type(delta) is int and delta >= 0
        assert type(cmvn) is bool
        return {
            'feat_type': feat_type,
            'channel': channel,
            'log': log,
            'delta': delta,
            'cmvn': cmvn
        }

    def forward(self, wavs=None, feat_list=None, wavs_len=None):
        # wavs: (*, channel_size, max_len)
        # feat_list, mam_list: [{feat_type: 'mfcc', channel: 0, log: False, delta: 2, cmvn: 'True'}, ...]
        # wavs_len: [len1, len2, ...]

        feat_list = self._check_list(feat_list)
        if wavs is None:
            max_channel_id = max([int(args['channel']) if 'channel' in args else 0 for args in feat_list])
            wavs = self._pseudo_wavs[0].view(1, 1, -1).repeat(1, max_channel_id + 1, 1)
        assert wavs.dim() >= 3

        # find length of wavs from padded tensor
        if wavs_len is None:
            wavs_len = []
            for wav in wavs:
                nonzero_index = wav.nonzero()
                if len(nonzero_index) == 0:
                    wavs_len.append(wav.size(-1)) # when all elements are zero
                else:
                    wavs_len.append(nonzero_index[:, -1].max().item()+1)
            wavs = pad_sequence([wav[:wav_len].transpose(-1, -2) for wav, wav_len in zip(wavs, wavs_len)], batch_first=True).transpose(-1, -2)

        wav = wavs.unsqueeze(2)
        shape = wavs.size()
        complx = self._stft(wavs.reshape(-1, shape[-1]), window=self._window)
        complx = complx.reshape(shape[:-1] + complx.shape[-3:])
        # complx: (*, channel_size, feat_dim, max_len, 2)
        linear, phase = self._magphase(complx)
        mel = self._melscale(linear)
        mfcc = self._mfcc_trans(wavs)
        complx = complx.transpose(-1, -2).reshape(*mfcc.shape[:2], -1, mfcc.size(-1))
        # complx, linear, phase, mel, mfcc: (*, channel_size, feat_dim, max_len)

        def select_feat(variables, feat_type, channel=0, log=False, delta=0, cmvn=False):
            raw_feat = variables[feat_type].select(dim=-3, index=channel)
            # apply log scale
            if bool(log):
                raw_feat = (raw_feat + self.eps).log()   
            feats = [raw_feat.contiguous()]
            # apply delta for features
            for _ in range(int(delta)):
                feats.append(compute_deltas(feats[-1]))
            feats = torch.cat(feats, dim=-2)
            downsample_rate = wavs.size(-1) / feats.size(-1)
            feats_len = [round(length / downsample_rate) for length in wavs_len]
            # apply cmvn
            if bool(cmvn):
                cmvn_feats = []
                for feat, feat_len in zip(feats, feats_len):
                    feat = feat[:, :feat_len]
                    cmvn_feat = (feat - feat.mean(dim=-1, keepdim=True)) / (feat.std(dim=-1, keepdim=True) + self.eps)
                    cmvn_feats.append(cmvn_feat.transpose(-1, -2))
                feats = pad_sequence(cmvn_feats, batch_first=True).transpose(-1, -2)
            return feats
            # return: (*, feat_dim, max_len)
        
        local_variables = locals()
        return self._transpose_list([select_feat(local_variables, **args) for args in feat_list])
        # return: [(*, max_len, feat_dim), ...]

    def istft(self, linears=None, phases=None, linear_power=2, complxs=None):
        assert complxs is not None or (linears is not None and phases is not None)
        # complxs: (*, n_freq, max_feat_len, 2) or (*, max_feat_len, n_freq * 2)
        # linears, phases: (*, max_feat_len, n_freq)

        if complxs is None:
            linears, phases = self._transpose_list([linears, phases])
            complxs = linears.pow(1/linear_power).unsqueeze(-1) * torch.stack([phases.cos(), phases.sin()], dim=-1)
        if complxs.size(-1) != 2:
            # treat complxs as: (*, max_feat_len, n_freq * 2)
            shape = complxs.size()
            complxs = complxs.view(*shape[:-1], -1, 2).transpose(-2, -3).contiguous()
        # complxs: (*, n_freq, max_feat_len, 2)
        
        return self._istft(complxs, window=self._window)
        # return: (*, max_wav_len)

    def test_istft(self, wavs=None, atol=1e-6):
        # wavs: (*, channel_size, max_wav_len)
        channel1, channel2 = 0, 1
        max_channel_id = max(channel1, channel2)
        
        if wavs is None:    
            wavs = self._pseudo_wavs[:max_channel_id + 1].unsqueeze(0)
        assert wavs.size(-2) > max_channel_id
        
        feat_list = [
            {'feat_type': 'complx', 'channel': channel1},
            {'feat_type': 'linear', 'channel': channel2},
            {'feat_type': 'phase', 'channel': channel2}
        ]
        complxs, linears, phases = self.forward(wavs, feat_list)
        assert torch.allclose(wavs.select(dim=-2, index=channel1), self.istft(complxs=complxs), atol=atol)
        assert torch.allclose(wavs.select(dim=-2, index=channel2), self.istft(linears=linears, phases=phases), atol=atol)
        print('[Preprocessor] test passed: stft -> istft')



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ dataset.py ]
#   Synopsis     [ the dataset that applies the spec augment pre-training task ]
#   Author       [ Andy T. Liu (https://github.com/andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import os
import numpy as np
#-------------#
import torch
from torch.nn.utils.rnn import pad_sequence
import torchaudio
#-------------#
from pretrain.mockingjay.dataset import KaldiAcousticDataset as _KaldiAcousticDataset
from pretrain.mockingjay.dataset import OnlineAcousticDataset as _OnlineAcousticDataset
from pretrain.spec_augment.task import generate_spec_aug_data


HALF_BATCHSIZE_TIME = 99999


class KaldiAcousticDataset(_KaldiAcousticDataset):
    
    def __init__(self, extracter, task_config, bucket_size, file_path, sets, 
                 max_timestep=0, libri_root=None, **kwargs):
        super(KaldiAcousticDataset, self).__init__(extracter, task_config, bucket_size, file_path, sets, 
                                                   max_timestep, libri_root, **kwargs)

    def __getitem__(self, index):
        # Load acoustic feature and pad
        x_batch = [self._sample(self._load_feat(x_file)) for x_file in self.X[index]]
        x_pad_batch = pad_sequence(x_batch, batch_first=True)
        return generate_spec_aug_data(spec=(x_pad_batch,), config=self.task_config)


class OnlineAcousticDataset(_OnlineAcousticDataset):
    
    def __init__(self, extracter, task_config, bucket_size, file_path, sets, 
                 max_timestep=0, libri_root=None, target_level=-25, **kwargs):
        super(OnlineAcousticDataset, self).__init__(extracter, task_config, bucket_size, file_path, sets, 
                                                    max_timestep, libri_root, target_level, **kwargs)
  
    def _process_x_pad_batch(self, x_pad_batch):
        if self.libri_root is not None:
            x_pad_batch = x_pad_batch.unsqueeze(1) # (batch_size, channel=1, seq_len)
            feat_list = self.extracter(x_pad_batch)
        return generate_spec_aug_data(feat_list, config=self.task_config)



import shutil
import random
import tempfile
from pathlib import Path
from typing import Any

import torch
import torchaudio
from torch.nn.utils.rnn import pad_sequence

SAMPLE_RATE = 16000


class pseudo_audio:
    def __init__(self, secs: int, sample_rate: int = SAMPLE_RATE):
        self.tempdir = Path(tempfile.TemporaryDirectory().name)
        self.tempdir.mkdir(parents=True, exist_ok=True)
        self.num_samples = []
        for n, sec in enumerate(secs):
            wav = torch.randn(1, sample_rate * sec)
            torchaudio.save(
                str(self.tempdir / f"{n}.wav"), wav, sample_rate=sample_rate
            )
            self.num_samples.append(wav.size(-1))
        self.filepaths = [str(self.tempdir / f"{i}.wav") for i in range(len(secs))]

    def __enter__(self):
        return self.filepaths, self.num_samples

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        shutil.rmtree(self.tempdir)


def get_pseudo_wavs(
    seed: int = 0,
    n: int = 2,
    min_secs: int = 1,
    max_secs: int = 3,
    sample_rate: int = SAMPLE_RATE,
    device: str = "cpu",
    padded: bool = False,
):
    random.seed(seed)
    torch.manual_seed(seed)

    wavs = []
    wavs_len = []
    for _ in range(n):
        wav_length = random.randint(min_secs * sample_rate, max_secs * sample_rate)
        wav = torch.randn(wav_length, requires_grad=True).to(device)
        wavs_len.append(wav_length)
        wavs.append(wav)

    if not padded:
        return wavs
    else:
        return pad_sequence(wavs, batch_first=True), torch.LongTensor(wavs_len)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/tera/hubconf.py ]
#   Synopsis     [ the tera torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def tera_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
        feature_selection (int): -1 (default, the last layer) or an int in range(0, max_layer_num)
    """
    assert os.path.isfile(ckpt)
    return _UpstreamExpert(ckpt, *args, **kwargs)


def tera_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return tera_local(_urls_to_filepaths(ckpt, refresh=refresh), *args, **kwargs)


def tera(refresh=False, *args, **kwargs):
    """
    The default model
        refresh (bool): whether to download ckpt/config again if existed
    """
    return tera_960hr(refresh, *args, **kwargs)


###########
# ALIASES #
###########


def tera_100hr(refresh=False, *args, **kwargs):
    """
    The tera base model on 100hr
        refresh (bool): whether to download ckpt/config again if existed
    """
    return tera_logMelBase_T_F_M_AdamW_b32_200k_100hr(refresh, *args, **kwargs)


def tera_960hr(refresh=False, *args, **kwargs):
    """
    The tera base model on 960hr
        refresh (bool): whether to download ckpt/config again if existed
    """
    return tera_logMelBase_T_F_M_AdamW_b32_1m_960hr_drop1(refresh, *args, **kwargs)


##########
# 100 HR #
##########


def tera_logMelBase_T_F_AdamW_b32_200k_100hr(refresh=False, *args, **kwargs):
    """
        Feature: 80-dim log Mel
        Alteration: time + freq
        Optimizer: AdamW
        Batch size: 32
        Total steps: 200k
        Unlabled Speech: 100hr
    """
    kwargs["ckpt"] = "https://www.dropbox.com/s/o36qt1zgtn3tsep/states-200000.ckpt?dl=1"
    return tera_url(refresh=refresh, *args, **kwargs)


def tera_logMelBase_T_F_M_AdamW_b32_200k_100hr(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time + freq + mag
    Optimizer: AdamW
    Batch size: 32
    Total steps: 200k
    Unlabled Speech: 100hr
    """
    kwargs["ckpt"] = "https://www.dropbox.com/s/l9ryl82k64m1lsk/states-200000.ckpt?dl=1"
    return tera_url(refresh=refresh, *args, **kwargs)


##########
# 960 HR #
##########


def tera_logMelBase_T_F_AdamW_b32_1m_960hr(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time + freq
    Optimizer: AdamW
    Batch size: 32
    Total steps: 1M
    Unlabled Speech: 960hr
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/98olxex0m7oy9ta/states-1000000.ckpt?dl=1"
    return tera_url(refresh=refresh, *args, **kwargs)


def tera_logMelBase_T_F_AdamW_b32_1m_960hr_drop1(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time + freq
    Optimizer: AdamW
    Batch size: 32
    Total steps: 1M
    Unlabled Speech: 960hr
    Differences: Dropout of 0.1 (instead of 0.3)
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/2ekbt2gxlkbvfz0/states-1000000.ckpt?dl=1"
    return tera_url(refresh=refresh, *args, **kwargs)


def tera_logMelBase_T_F_AdamW_b32_1m_960hr_seq3k(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time + freq
    Optimizer: AdamW
    Batch size: 32
    Total steps: 1M
    Unlabled Speech: 960hr
    Differences: sequence length of 3k (instead of 1.5k)
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/tfysinbalpm3gsj/states-1000000.ckpt?dl=1"
    return tera_url(refresh=refresh, *args, **kwargs)


def tera_logMelBase_T_F_M_AdamW_b32_1m_960hr_drop1(refresh=False, *args, **kwargs):
    """
    Feature: 80-dim log Mel
    Alteration: time + freq + mag
    Optimizer: AdamW
    Batch size: 32
    Total steps: 1M
    Unlabled Speech: 960hr
    Differences: Dropout of 0.1 (instead of 0.3)
    """
    kwargs[
        "ckpt"
    ] = "https://www.dropbox.com/s/xdoj9wdo87lztv1/states-1000000.ckpt?dl=1"
    return tera_url(refresh=refresh, *args, **kwargs)


#####################
# Other Feat 100 HR #
#####################


def tera_fbankBase_T_F_AdamW_b32_200k_100hr(refresh=False, *args, **kwargs):
    """
    Feature: 240-dim fbank
    Alteration: time + freq
    Optimizer: AdamW
    Batch size: 32
    Total steps: 200k
    Unlabled Speech: 100hr
    """
    kwargs["ckpt"] = "https://www.dropbox.com/s/i32ob29m6afufot/states-200000.ckpt?dl=1"
    return tera_url(refresh=refresh, *args, **kwargs)



import argparse
import torch

import hubconf


parser = argparse.ArgumentParser()
upstreams = [attr for attr in dir(hubconf) if callable(getattr(hubconf, attr)) and attr[0] != '_']
parser.add_argument('--mode', choices=['list', 'help', 'load'], required=True)
parser.add_argument('--upstream', choices=upstreams)
parser.add_argument('--ckpt', help='The PATH/URL/GOOGLE_DRIVE_ID of upstream checkpoint, not always needed')
parser.add_argument('--config', help='The PATH of upstream config, not always needed')
parser.add_argument('--refresh', action='store_true', help='Whether to re-download upstream contents')

args = parser.parse_args()

if args.mode == 'list':
    print(torch.hub.list('s3prl/s3prl', force_reload=args.refresh))

elif args.mode == 'help':
    print(torch.hub.help('s3prl/s3prl', args.upstream, force_reload=args.refresh))

elif args.mode == 'load':
    print(torch.hub.load(
        's3prl/s3prl', args.upstream, force_reload=args.refresh,
        ckpt=args.ckpt, config=args.config, refresh=args.refresh
    ))



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/cpc/expert.py ]
#   Synopsis     [ the cpc wrapper ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import torch
import argparse

import torch
from torch.nn.utils.rnn import pad_sequence

from ..interfaces import UpstreamBase
from .model import CPCModel as cpcmodel
from .cpc_default_config import get_default_cpc_config
from .feature_loader import getEncoder, getAR, loadArgs

SAMPLE_RATE = 16000
EXAMPLE_SEC = 3
EXAMPLE_BATCH_SIZE = 32


class UpstreamExpert(UpstreamBase):
    def __init__(self, ckpt, **kwargs):
        super().__init__(**kwargs)

        locArgs = get_default_cpc_config()
        checkpoint = torch.load(ckpt, map_location="cpu")
        loadArgs(locArgs, argparse.Namespace(**checkpoint["config"]))

        encoderNet = getEncoder(locArgs)
        arNet = getAR(locArgs)
        self.model = cpcmodel(encoderNet, arNet)
        self.model.load_state_dict(checkpoint["weights"], strict=False)

        if len(self.hooks) == 0:
            self.add_hook(
                "self.model.gEncoder", lambda input, output: output.transpose(1, 2)
            )
            self.add_hook("self.model.gAR", lambda input, output: output)

    def get_downsample_rates(self, key: str) -> int:
        return 160

    def forward(self, wavs):
        padded_wav = pad_sequence(wavs, batch_first=True)
        features = self.model(padded_wav.unsqueeze(1), None)[0]

        # This forward function only does the model forward
        # The return dict is then handled by UpstreamBase's hooks



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/mockingjay/builder.py ]
#   Synopsis     [ build the transformer model for downstream usage ]
#   Author       [ Andy T. Liu (https://github.com/andi611) ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import sys
import copy
import math
import yaml
import torch
import random
import torchaudio
import numpy as np
import torch.nn as nn
from torch.nn.utils.rnn import pad_sequence
from functools import lru_cache
from distutils.util import strtobool
from ..baseline.extracter import get_extracter
from ..baseline.preprocessor import get_preprocessor
from .model import TransformerConfig, TransformerModel
from .model import TransformerSpecPredictionHead
import s3prl.optimizers

#######################
# TRANSFORMER BUILDER #
#######################
class TransformerBuilder(nn.Module):
    """ 
    A builder class for all pre-trained Transformer.
    Child classes only need to implement the __init__() and forward() method.
    """
    def __init__(self, options, inp_dim=-1, config=None, on_the_fly_config=None, verbose=False):
        super(TransformerBuilder, self).__init__()

        # read config
        if config is not None:
            self.config = yaml.load(open(config, 'r'), Loader=yaml.FullLoader)
        else:
            # Since some old checkpoints contained pickled scheduler which needs 'optimizers'
            # module which is now moved into s3prl package.
            original_optimizer = sys.modules.get("optimizers")
            sys.modules["optimizers"] = s3prl.optimizers

            self.all_states = torch.load(options["ckpt_file"], map_location='cpu')
            if 'transformer' in self.all_states['Config']:
                # for legacy ckpts, 'Config' stored the upstream config, and runner config was stored under 'Runner'.
                self.config = self.all_states['Config']
            elif 'transformer' in self.all_states['Upstream_Config']:
                # For new ckpts, 'Upstream_Config' is used to store the upstream config, and runner config is now stored under 'Config'.
                self.config = self.all_states['Upstream_Config']
            else:
                raise NotImplementedError

            del sys.modules["optimizers"]
            if original_optimizer is not None:
                sys.modules["optimizers"] = original_optimizer

        # parse the options dict
        self.load = bool(strtobool(options["load_pretrain"]))
        self.no_grad = bool(strtobool(options['no_grad']))
        self.spec_aug = bool(strtobool(options['spec_aug']))
        self.spec_aug_prev = bool(strtobool(options['spec_aug_prev']))
        self.output_hidden_states = bool(strtobool(options['output_hidden_states']))
        self.select_layer = int(options['select_layer'])
        self.permute_input = bool(strtobool(options['permute_input']))
        if (not self.no_grad) and (not self.spec_aug_prev): raise RuntimeError('Only one of them can be set False!')
        if str(options['dropout']) != 'default': # increase dropout if specified
            self.config['transformer']['hidden_dropout_prob'] = float(options['dropout'])
            self.config['transformer']['attention_probs_dropout_prob'] = float(options['dropout'])

        # Set model config
        self.model_config = TransformerConfig(self.config['transformer'])
        self.hidden_size = self.model_config.hidden_size
        self.num_layers = self.model_config.num_hidden_layers
        self.max_input_length = self.config['task']['sequence_length']

        if on_the_fly_config is not None:
            self.config['audio'] = yaml.load(open(on_the_fly_config, 'r'), Loader=yaml.FullLoader)
        if 'audio' in self.config:
            if 'kaldi' in self.config['audio']:
                self.extracter, self.inp_dim, _ = get_extracter(self.config['audio'])
                self.spec_dim = self.inp_dim
            else:
                self.extracter, self.inp_dim, self.spec_dim = get_preprocessor(self.config['audio'], process_input_only=True)
                self.target_level = self.config['audio']['target_level']
        elif inp_dim != -1:
            self.extracter, self.inp_dim, self.spec_dim = None, inp_dim, inp_dim
        else:
            self.extracter, self.inp_dim, self.spec_dim = None, self.config['transformer']['input_dim'], self.config['transformer']['input_dim']
        
        if self.max_input_length > 0 and verbose: print('[Transformer] - Maximum input length: ', self.max_input_length)
        if not (self.select_layer in list(range(-1, self.num_layers))): raise RuntimeError('Out of range int for \'select_layer\'!')


    def _normalize_wav_decibel(self, wav):
        '''Normalize the signal to the target level'''
        rms = wav.pow(2).mean().pow(0.5)
        scalar = (10 ** (self.target_level / 20)) / (rms + 1e-10)
        wav = wav * scalar
        return wav


    def load_model(self, transformer_model, state_dict, verbose=False):
        try:
            old_keys = []
            new_keys = []
            for key in state_dict.keys():
                new_key = None
                if 'gamma' in key:
                    new_key = key.replace('gamma', 'weight')
                if 'beta' in key:
                    new_key = key.replace('beta', 'bias')
                if new_key:
                    old_keys.append(key)
                    new_keys.append(new_key)
            for old_key, new_key in zip(old_keys, new_keys):
                state_dict[new_key] = state_dict.pop(old_key)

            missing_keys = []
            unexpected_keys = []
            error_msgs = []
            # copy state_dict so _load_from_state_dict can modify it
            metadata = getattr(state_dict, '_metadata', None)
            state_dict = state_dict.copy()
            if metadata is not None:
                state_dict._metadata = metadata

            def load(module, prefix=''):
                local_metadata = {} if metadata is None else metadata.get(prefix[:-1], {})
                module._load_from_state_dict(
                    state_dict, prefix, local_metadata, True, missing_keys, unexpected_keys, error_msgs)
                for name, child in module._modules.items():
                    if child is not None:
                        load(child, prefix + name + '.')

            load(transformer_model)
            if len(missing_keys) > 0:
                print('Weights of {} not initialized from pretrained model: {}'.format(
                    transformer_model.__class__.__name__, missing_keys))
            if len(unexpected_keys) > 0:
                print('Weights from pretrained model not used in {}: {}'.format(
                    transformer_model.__class__.__name__, unexpected_keys))
            if len(error_msgs) > 0:
                raise RuntimeError('Error(s) in loading state_dict for {}:\n\t{}'.format(
                                    transformer_model.__class__.__name__, '\n\t'.join(error_msgs)))
            if verbose: print('[Transformer] - Pre-trained weights loaded!')
            return transformer_model

        except: 
            raise RuntimeError('[Transformer] - Pre-trained weights NOT loaded!')
        

    def process_input_data(self, feat):
        """Process input data for the model"""
        
        # add arbitary batch axis B if input `feat` has shape of TxD
        if len(feat.shape) == 2:
            feat = feat.unsqueeze(0)
        # input `feat` should have shape BxTxD
        elif len(feat.shape) != 3:
            raise ValueError('Input argument `feat` has invalid shape: {}'.format(feat.shape))

        # Record length for each uttr
        spec_len = (feat.sum(dim=-1) != 0).long().sum(dim=-1)
        spec_len = spec_len.tolist()

        batch_size = feat.shape[0]
        seq_len = feat.shape[1]

        pos_enc = position_encoding(seq_len, self.hidden_size) # (seq_len, hidden_size)
        attn_mask = np.ones((batch_size, seq_len)) # (batch_size, seq_len)

        # zero vectors for padding dimension
        for idx in range(len(feat)):
            attn_mask[idx][spec_len[idx]:] = 0 

        if self.spec_aug and self.spec_aug_prev and self.model.training and self.inp_dim > 1:
            feat = spec_augment(feat, mask_T=70, mask_F=9, num_T=2, num_F=2, p=1.0) # (batch_size, seq_len, feature_dim * dr)
        feat = feat.to(dtype=torch.float32) # (batch_size, seq_len, feature_dim * dr)
        pos_enc = torch.FloatTensor(pos_enc).to(device=feat.device, dtype=torch.float32).expand(feat.size(0), *pos_enc.size()) # (batch_size, seq_len, hidden_size)
        attn_mask = torch.FloatTensor(attn_mask).to(device=feat.device, dtype=torch.float32) # (batch_size, seq_len)
        return feat, pos_enc, attn_mask # (x, pos_enc, attention_mask)


    def _forward(self, x):

        if self.permute_input:
            x = x.permute(1, 0, 2).contiguous() # (T, B, D) -> (B, T, D)
        input_len = x.shape[1]

        # forward the whole sequence at once
        if self.max_input_length == 0 or input_len <= self.max_input_length:
            feat, pos_enc, attn_mask = self.process_input_data(x) # x shape: (B, T, D)
            x = self.model(feat, pos_enc, attn_mask, output_all_encoded_layers=self.output_hidden_states or self.select_layer != -1)
            x = torch.stack(x) if isinstance(x, list) else x  # (B, T, D) or # (N, B, T, D)
        # forward the sequence in chunks then concat
        else:
            chunks = torch.chunk(x, chunks=math.ceil(input_len / self.max_input_length), dim=1)
            x_ = []
            for chunk in chunks:
                feat, pos_enc, attn_mask = self.process_input_data(chunk) # x shape: (B, T, D)
                chunk = self.model(feat, pos_enc, attn_mask, output_all_encoded_layers=self.output_hidden_states or self.select_layer != -1) # (B, T, D) or # (N, B, T, D)
                x_.append(torch.stack(chunk) if type(chunk) is list else chunk)
            x = torch.cat(x_, dim=2 if (self.output_hidden_states or self.select_layer != -1) else 1)
        # x can be a single hidden state or a list of hidden states

        if self.output_hidden_states:
            hidden_states = x
            x = hidden_states[self.select_layer]

        if self.spec_aug and not self.spec_aug_prev and self.model.training and self.inp_dim > 1:
            x = spec_augment(x, mask_T=70, mask_F=86, num_T=2, num_F=2, p=1.0) # (B, T, D)
        
        # permute to output
        if self.permute_input:
            x = x.permute(1, 0, 2).contiguous() # (B, T, D) -> (T, B, D)
            if self.output_hidden_states:
                # (N, B, T, D) -> (N, T, B, D)
                hidden_states = hidden_states.transpose(1, 2).contiguous()

        if self.output_hidden_states:
            return x, hidden_states
        return x # (B, T, D) or (T, B, D)


##########################
# PRETRAINED TRANSFORMER #
##########################
class PretrainedTransformer(TransformerBuilder):
    """
    Use this class to extract features from the Transformer model,
    or to finetune the pre-trained Transformer with any downstream tasks.
    Also, this class is `pytorch-kaldi` ready,
    hence we need to use `str` instead of `bool` in the options dict,
    as pytorch-kaldi scripts will pass in str.

    Params:
        `options`: a python dictionary containing the following keys:
            ckpt_file: str, a path specifying the pre-trained ckpt file
            load_pretrain: str, ['True', 'False'], whether to load pre-trained weights
            no_grad: str, ['True', 'False'], whether to use torch.no_grad over forward, this determines should torch build the computational graph
            dropout: float/str, use float to modify dropout value during downstream finetune, or use the str `default` for pre-train default values
            spec_aug: str, ['True', 'False'], whether to apply the SpecAugment technique
            spec_aug_prev: str, ['True', 'False'], True: apply spec augment on input (i.e. acoustic features); False: apply on output (i.e. the hidden states)
            weighted_sum: str, ['True', 'False'], whether to use a learnable weighted sum to integrate hidden representations from all layers, if False then use the one specified in `select_layer`
            select_layer: int, select from all hidden representations, set to -1 to select the last (will only be used when weighted_sum is False)
            permute_input: str, ['True', 'False'], this attribute is for the forward method. Ture: input / ouput of shape (T, B, D); False: input / ouput of shape (B, T, D)
        `intput_dim`: int, input dimension of model
        `config`: optional, reads the given yaml config and not use the config stored in `ckpt_file`

    An example `options` dictionary:
    options = {
        'ckpt_file'     : './result/result_transformer/libri_sd1337_fmllrBase960-F-N-K-RA/states-1000000.ckpt',
        'load_pretrain' : 'True',
        'no_grad'       : 'True',
        'dropout'       : 'default',
        'spec_aug'      : 'False',
        'spec_aug_prev' : 'True',
        'weighted_sum'  : 'False',
        'select_layer'  : -1,
        'permute_input' : 'False',
    }
    """
    def __init__(self, options, inp_dim, config=None, online_config=None, verbose=False):
        super(PretrainedTransformer, self).__init__(options, inp_dim, config, online_config, verbose)

        # Build model
        self.model = TransformerModel(self.model_config, self.inp_dim)
        self.model.eval() if self.no_grad else self.model.train()
        self.out_dim = self.hidden_size # This attribute is necessary, for pytorch-kaldi and run_downstream.py
        
        # Load from a PyTorch state_dict
        if self.load: 
            self.model = self.load_model(self.model, self.all_states['Transformer'], verbose)
            if verbose: print('[Transformer] - Number of parameters: ' + str(sum(p.numel() for p in self.model.parameters() if p.requires_grad)))


    def forward(self, x):
        if self.extracter is not None:
            if 'kaldi' in self.config['audio']:
                x = [self.extracter(x_i) for x_i in x]
                x = pad_sequence(x, batch_first=True)
            else:
                x = [self._normalize_wav_decibel(x_i) for x_i in x]
                x_lens = [len(x_) for x_ in x]
                x = pad_sequence(x, batch_first=True)
                x = x.unsqueeze(1) # (batch_size, audio_len) -> (batch_size, 1, audio_len)
                x = self.extracter(x, wavs_len=x_lens)[0]
        if self.no_grad:
            with torch.no_grad():
                x = self._forward(x)
        else:
            x = self._forward(x)
        return x


####################################
# PRETRAINED TRANSFORMER WITH HEAD #
####################################
class PretrainedTransformerWithHead(PretrainedTransformer):
    def __init__(self, options, inp_dim, config=None, online_config=None, verbose=False):
        super(PretrainedTransformerWithHead, self).__init__(options, inp_dim, config, online_config, verbose)

        # build head
        self.SpecHead = TransformerSpecPredictionHead(self.model_config, self.spec_dim)
        self.SpecHead.eval() if self.no_grad else self.SpecHead.train()
        
        # Load from a PyTorch state_dict
        if self.load:
            self.SpecHead.load_state_dict(self.all_states['SpecHead'])
            if verbose: print('[Spec Transformer] - Number of parameters: ' + str(sum(p.numel() for p in self.SpecHead.parameters() if p.requires_grad)))

    def forward(self, x):
        if self.extracter is not None:
            if 'kaldi' in self.config['audio']:
                x = [self.extracter(x_i) for x_i in x]
                x = pad_sequence(x, batch_first=True)
            else:
                x = [self._normalize_wav_decibel(x_i) for x_i in x]
                x_lens = [len(x_) for x_ in x]
                x = pad_sequence(x, batch_first=True)
                x = x.unsqueeze(1) # (batch_size, audio_len) -> (batch_size, 1, audio_len)
                x = self.extracter(x, wavs_len=x_lens)[0]
        if self.no_grad:
            with torch.no_grad():
                x = self._forward(x)
                x, _ = self.SpecHead(x)
        else:
            x = self._forward(x)
            x, _ = self.SpecHead(x)
        return x


#######################
# POSITIONAL ENCODING #
#######################
MAX_SEQLEN = 24000
@lru_cache(maxsize=128)
def get_sinusoid_table(hidden_size):
    def _cal_angle(position, hid_idx):
        return position / np.power(10000, 2 * (hid_idx // 2) / hidden_size)
    def _get_posi_angle_vec(position):
        return [_cal_angle(position, hid_j) for hid_j in range(hidden_size)]
    sinusoid_table = np.array([_get_posi_angle_vec(pos_i) for pos_i in range(MAX_SEQLEN)])
    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1
    return torch.FloatTensor(sinusoid_table)


def position_encoding(seq_len, hidden_size):
    """ position encoding table """     
    table = get_sinusoid_table(hidden_size)[:seq_len]
    # no extra CPU and GPU memory allocation
    # after getting the (seq_len, hidden_size) tensor, one should first put
    # this tensor into GPU then expand it
    return table  # (seq_len, hidden_size)


################
# SPEC AUGMENT #
################
"""
Process training data for the supervised ASR model by 
masking to time-steps and channels during training
which delays overfitting and significantly improves the final accuracy numbers.
Input:
    `spec`: input real frames, with shape: (batch_size, seq_len, feature_dim)
    `mask_T`: the time mask parameter T described in the SpecAugment paper, 
              we use default values based on the LD Policy
              (In paper: T=100, we use 70 since we are training on the 100 hr subset only)
    `mask_F`: the frequency mask parameter F described in the SpecAugment paper, 
              we use default values based on the LD Policy
              (In paper: F=27:D=80*3 -> F=9:D=80, where D is acoustic dimension)
    `num_T` : the number of time masks applied (In paper: mT=2)
    `num_F` : the number of frequency masks applied (In paper: mF=2)
    `p` : upper bound ratio (In paper: p=1.0)
Output:
    `spec`: augmented frames, with shape: (batch_size, seq_len, feature_dim)
"""
def spec_augment(spec, mask_T=70, mask_F=9, num_T=2, num_F=2, p=1.0):

    def _start_to_intervals(starts, consecutive):
        tiled = starts.expand(consecutive, starts.size(0)).permute(1, 0)
        offset = torch.arange(consecutive).expand_as(tiled)
        intervals = tiled + offset
        return intervals.view(-1)

    with torch.no_grad():
        upper_bound = spec.shape[1] * p # upper bound on the time mask so that a time mask cannot be wider than p times the number of time steps
        
        for idx in range(spec.shape[0]):

            # time masking
            if mask_T > 0 and mask_T < upper_bound:
                for _ in range(num_T):
                    rand_consecutive = random.randint(0, mask_T)
                    chosen_start = torch.randperm(spec.shape[1] - rand_consecutive)[:1]
                    chosen_intervals = _start_to_intervals(chosen_start, rand_consecutive)
                    spec[idx, chosen_intervals, :] = 0

            # frequency masking
            if mask_F > 0:
                for _ in range(num_F):
                    rand_bandwidth = random.randint(0, mask_F)
                    chosen_start = torch.randperm(spec.shape[2] - rand_bandwidth)[:1]
                    chosen_intervals = _start_to_intervals(chosen_start, rand_bandwidth)
                    spec[idx, :, chosen_intervals] = 0

        return spec


#######
# LIN #
#######
"""
Linear Input Networks (LIN) for domain adaptation
Params:
    `options`: a python dictionary containing arguments for pytorch kaldi, give None if not using with pytorch-kaldi:
    `intput_dim`: int, input dimension of model
"""
class LIN(nn.Module):
    def __init__(self, options, inp_dim):
        super(LIN, self).__init__()

        self.out_dim = inp_dim # This attribute is for pytorch-kaldi
        self.linear = nn.Linear(inp_dim, inp_dim)
        self.linear.weight.data.copy_(torch.eye(inp_dim))
        
        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        self.linear = self.linear.to(self.device)
        self.linear.train()
        
    def forward(self, x):
        x = self.linear(x)
        return x



import logging
from typing import List

import torch
import torch.nn as nn
import torch.nn.functional as F

from s3prl import hub
from s3prl.util.pseudo_data import get_pseudo_wavs

logger = logging.getLogger(__name__)

CHECK_ITERATION = 10
TOLERABLE_SEQLEN_DIFF = 3


class S3PRLUpstream(nn.Module):
    """
    This is an easy interface for using all the models in S3PRL.

    Args:
        name (str):
            can be "apc", "hubert", "wav2vec2". See :obj:`available_names` for all the supported names

        path_or_url (str):
            The source of the checkpoint. Might be a local path or a URL

        refresh (bool): (default, False)
            If false, only downlaod checkpoint if not yet downloaded before.
            If true, force to re-download the checkpoint.

    .. note::

        When using **S3PRLUpstream** with :code:`refresh=True` and multiprocessing (e.g. DDP),
        the checkpoint will only be downloaded once, and the other processes will simply
        re-use the newly downloaded checkpoint, instead of re-downloading on every processes,
        which can be very time/bandwidth consuming.

    Example::

        >>> import torch
        >>> from s3prl.nn import S3PRLUpstream
        ...
        >>> model = S3PRLUpstream("hubert")
        >>> model.eval()
        ...
        >>> with torch.no_grad():
        ...     wavs = torch.randn(2, 16000 * 2)
        ...     wavs_len = torch.LongTensor([16000 * 1, 16000 * 2])
        ...     all_hs, all_hs_len = model(wavs, wavs_len)
        ...
        >>> for hs, hs_len in zip(all_hs, all_hs_len):
        ...     assert isinstance(hs, torch.FloatTensor)
        ...     assert isinstance(hs_len, torch.LongTensor)
        ...
        ...     batch_size, max_seq_len, hidden_size = hs.shape
        ...     assert hs_len.dim() == 1
    """

    @classmethod
    def available_names(cls, only_registered_ckpt: bool = False) -> List[str]:
        """
        All the available names supported by this S3PRLUpstream

        Args:
            only_registered_ckpt (bool):
                ignore entry names which require to give `path_or_url`.
                That is, the entry names without the registered checkpoint sources.
                These names end with :code:`_local` (for local path), :code:`_url`
                (for URL) or :code:`_custom` (auto-determine path or URL)
        """
        return hub.options(only_registered_ckpt)

    def __init__(self, name: str, path_or_url: str = None, refresh: bool = False):
        super().__init__()
        self.upstream = getattr(hub, name)(ckpt=path_or_url, refresh=refresh)

        self.upstream.eval()
        with torch.no_grad():
            hs = self.upstream(get_pseudo_wavs())["hidden_states"]
        self.upstream.train()
        self._num_layers = len(hs)

        self._hidden_sizes = []
        for h in hs:
            self._hidden_sizes.append(h.size(-1))

        downsample_rates = self.upstream.get_downsample_rates("hidden_states")
        if isinstance(downsample_rates, int):
            self._downsample_rates = [downsample_rates] * self._num_layers
        elif isinstance(downsample_rates, (tuple, list)):
            self._downsample_rates = downsample_rates
        else:
            raise ValueError

    @property
    def num_layers(self) -> int:
        """
        Number of hidden sizes. All the upstream have a deterministic
        number of layers. That is, layer drop is turned off by default.
        """
        return self._num_layers

    @property
    def downsample_rates(self) -> List[int]:
        """
        Downsampling rate from 16000 Hz audio of each layer.
        Usually, all layers have the same downsampling rate,
        but might not be the case for some advanced upstreams.
        """
        return self._downsample_rates

    @property
    def hidden_sizes(self) -> List[int]:
        """
        The hidden size of each layer
        """
        return self._hidden_sizes

    def _match_length(self, xs, target_max_len: int):
        xs_max_len = xs.size(1)

        if xs_max_len > target_max_len:
            assert round(xs_max_len / target_max_len) == 1
            xs = xs[:, :target_max_len, :]

        elif xs_max_len < target_max_len:
            assert round(target_max_len / xs_max_len) == 1
            xs = torch.cat(
                (xs, xs[:, -1:, :].repeat(1, target_max_len - xs_max_len, 1)), dim=1
            )

        return xs

    def forward(self, wavs: torch.FloatTensor, wavs_len: torch.LongTensor):
        """
        Args:
            wavs (torch.FloatTensor): (batch_size, seqlen)
            wavs_len (torch.LongTensor): (batch_size, )

        Return:
            List[torch.FloatTensor], List[torch.LongTensor]

            1. all the layers of hidden states: List[ (batch_size, max_seq_len, hidden_size) ]
            2. the valid length for each hidden states: List[ (batch_size, ) ]
        """
        wavs_list = []
        for wav, wav_len in zip(wavs, wavs_len):
            wavs_list.append(wav[:wav_len])

        hidden_states = self.upstream(wavs_list)["hidden_states"]
        assert isinstance(hidden_states, (list, tuple))
        assert len(hidden_states) == self.num_layers

        max_wav_len = int(max(wavs_len))
        all_hs = []
        all_lens = []
        for h, stride in zip(hidden_states, self.downsample_rates):
            expected_max_h_len = max_wav_len // stride + 1
            h = self._match_length(h, expected_max_h_len)
            assert h.size(1) == expected_max_h_len
            all_hs.append(h)

            h_len = torch.div(wavs_len, stride, rounding_mode="floor") + 1
            all_lens.append(h_len)

        return all_hs, all_lens


class Featurizer(nn.Module):
    """
    Featurizer take the :obj:`S3PRLUpstream`'s multiple layer of hidden_states and
    reduce (standardize) them into a single hidden_states, to connect with downstream NNs.

    This basic Featurizer expects all the layers to have same stride and hidden_size
    When the input upstream only have a single layer of hidden states, use that directly.
    If multiple layers are presented, add a trainable weighted-sum on top of those layers.

    Args:
        upstream (:obj:`S3PRLUpstream`):
            the upstream to extract features, this upstream is used only for initialization
            and will not be kept in this Featurizer object
        layer_selections (List[int]):
            To select a subset of hidden states from the given upstream by layer ids (0-index)
            If None (default), than all the layer of hidden states are selected
        normalize (bool):
            Whether to apply layer norm on all the hidden states before weighted-sum
            This can help convergence in some cases, but not used in SUPERB to ensure the
            fidelity of each upstream's extracted representation.

    Example::

        >>> import torch
        >>> from s3prl.nn import S3PRLUpstream
        ...
        >>> model = S3PRLUpstream("hubert")
        >>> model.eval()
        ...
        >>> with torch.no_grad():
        ...     wavs = torch.randn(2, 16000 * 2)
        ...     wavs_len = torch.LongTensor([16000 * 1, 16000 * 2])
        ...     all_hs, all_hs_len = model(wavs, wavs_len)
        ...
        >>> featurizer = Featurizer(model)
        >>> hs, hs_len = featurizer(all_hs, all_hs_len)
        ...
        >>> assert isinstance(hs, torch.FloatTensor)
        >>> assert isinstance(hs_len, torch.LongTensor)
        >>> batch_size, max_seq_len, hidden_size = hs.shape
        >>> assert hs_len.dim() == 1
    """

    def __init__(
        self,
        upstream: S3PRLUpstream,
        layer_selections: List[int] = None,
        normalize: bool = False,
    ):
        super().__init__()
        assert len(set(upstream.hidden_sizes)) == 1
        assert len(set(upstream.downsample_rates)) == 1
        self._output_size = upstream.hidden_sizes[0]
        self._downsample_rate = upstream.downsample_rates[0]
        self.normalize = normalize

        if upstream.num_layers > 1:
            if layer_selections is not None:
                assert upstream.num_layers >= len(layer_selections)
                self.layer_selections = sorted(layer_selections)
            else:
                self.layer_selections = list(range(upstream.num_layers))
            self.weights = nn.Parameter(torch.zeros(len(self.layer_selections)))

    @property
    def output_size(self) -> int:
        """
        The hidden size of the final weighted-sum output
        """
        return self._output_size

    @property
    def downsample_rate(self) -> int:
        """
        The downsample rate (from 16k Hz waveform) of the final weighted-sum output
        """
        return self._downsample_rate

    def _weighted_sum(self, all_hs, all_lens):
        assert len(all_hs) == len(all_lens) > 1
        for l in all_lens[1:]:
            torch.allclose(all_lens[0], l)
        stacked_hs = torch.stack(all_hs, dim=0)

        if self.normalize:
            stacked_hs = F.layer_norm(stacked_hs, (stacked_hs.shape[-1],))

        _, *origin_shape = stacked_hs.shape
        stacked_hs = stacked_hs.view(len(self.layer_selections), -1)
        norm_weights = F.softmax(self.weights, dim=-1)
        weighted_hs = (norm_weights.unsqueeze(-1) * stacked_hs).sum(dim=0)
        weighted_hs = weighted_hs.view(*origin_shape)

        return weighted_hs, all_lens[0]

    def forward(
        self, all_hs: List[torch.FloatTensor], all_lens: List[torch.LongTensor]
    ):
        """
        Args:
            all_hs (List[torch.FloatTensor]): List[ (batch_size, seq_len, hidden_size) ]
            all_lens (List[torch.LongTensor]): List[ (batch_size, ) ]

        Return:
            torch.FloatTensor, torch.LongTensor

            1. The weighted-sum result, (batch_size, seq_len, hidden_size)
            2. the valid length of the result, (batch_size, )
        """
        if len(all_hs) == 1:
            return all_hs[0], all_lens[0]

        all_hs = [h for idx, h in enumerate(all_hs) if idx in self.layer_selections]
        all_lens = [l for idx, l in enumerate(all_lens) if idx in self.layer_selections]
        hs, hs_len = self._weighted_sum(all_hs, all_lens)
        return hs, hs_len



import copy
from unicodedata import bidirectional
import torch
import torch.nn as nn
from torch.nn.utils.rnn import pad_packed_sequence, pack_padded_sequence
#-------------#
class Decoar(nn.Module):
    def __init__(self):
        """
            input_size: an int indicating the input feature size, e.g., 80 for Mel.
            hidden_size: an int indicating the RNN hidden size.
            num_layers: an int indicating the number of RNN layers.
            dropout: a float indicating the RNN dropout rate.
            residual: a bool indicating whether to apply residual connections.
        """
        super(Decoar, self).__init__()

        self.embed = 80
        d = 1024
        self.encoder_layers = 4
        self.post_extract_proj = nn.Linear(self.embed, d)

        self.forward_lstms = nn.ModuleList([
            nn.LSTM(input_size=d, hidden_size=d, batch_first=True, bidirectional=False)
            for _ in range(self.encoder_layers)
        ])
        self.backward_lstms = nn.ModuleList([
            nn.LSTM(input_size=d, hidden_size=d, batch_first=True, bidirectional=False)
            for _ in range(self.encoder_layers)
        ])

    def flipBatch(self, data, lengths):
        assert data.shape[0] == len(lengths), "Dimension Mismatch!"
        for i in range(data.shape[0]):
            data[i, :lengths[i]] = data[i, :lengths[i]].flip(dims=[0])

        return data

    def forward(self, features, padding_mask=None):
        max_seq_len = features.shape[1]
        features = self.post_extract_proj(features)

        if padding_mask is not None:
            extra = padding_mask.size(1) % features.size(1)
            if extra > 0:
                padding_mask = padding_mask[:, :-extra]
            padding_mask = padding_mask.view(padding_mask.size(0), features.size(1), -1)
            padding_mask = padding_mask.all(-1)

        seq_lengths = (~padding_mask).sum(dim=-1).tolist()

        packed_rnn_inputs = pack_padded_sequence(features, seq_lengths,
                                                 batch_first=True,
                                                 enforce_sorted=False)

        forward_outputs = []
        packed_rnn_outputs = packed_rnn_inputs
        for forward_lstm in self.forward_lstms:
            packed_rnn_outputs, _ = forward_lstm(packed_rnn_outputs)
            forward_outputs.append(packed_rnn_outputs)

        packed_rnn_inputs = pack_padded_sequence(self.flipBatch(features, seq_lengths), seq_lengths,
                                                 batch_first=True,
                                                 enforce_sorted=False)
        
        backward_outputs = []
        packed_rnn_outputs = packed_rnn_inputs
        for backward_lstm in self.backward_lstms:
            packed_rnn_outputs, _ = backward_lstm(packed_rnn_outputs)
            backward_outputs.append(packed_rnn_outputs)

        concat_layer_output = []
        for forward_output, backward_output in zip(forward_outputs, backward_outputs):
            x_forward, _ = pad_packed_sequence(forward_output,
                                    batch_first=True,
                                    total_length=max_seq_len)
            x_backward, _ = pad_packed_sequence(backward_output,
                                    batch_first=True,
                                    total_length=max_seq_len)
            x_backward = self.flipBatch(x_backward, seq_lengths)
            concat_layer_output.append(torch.cat((x_forward, x_backward), dim=-1))

        return concat_layer_output
